from typing import Callable, Optional, Union

import keras
from keras import ops

from keras_rs.src import types


def get_shuffled_indices(
    shape: types.TensorShape,
    mask: Optional[types.Tensor] = None,
    shuffle_ties: bool = True,
    seed: Optional[Union[int, keras.random.SeedGenerator]] = None,
) -> types.Tensor:
    """Utility function for getting shuffled indices, with masked indices
    pushed to the end.

    Args:
        shape: tuple. The shape of the tensor for which to generate
            shuffled indices.
        mask: An optional boolean tensor with the same shape as `shape`.
            If provided, elements where `mask` is `False` will be placed
            at the end of the sorted indices. Defaults to `None` (no masking).
        shuffle_ties: Boolean indicating how to handle ties if multiple elements
            have the same sorting value (randomly when `shuffle_ties` is True
            otherwise, order is preserved).
        seed: Optional integer seed for the random number generator used when
            `shuffle_ties` is True. Ensures reproducibility. Defaults to None.

    Returns:
        A tensor of shape `shape` containing shuffled indices.
    """
    # If `shuffle_ties` is True, generate random values. Otherwise, generate
    # zeros so that we get `[0, 1, 2, ...]` as indices on doing `argsort`.
    if shuffle_ties:
        shuffle_values = keras.random.uniform(shape, seed=seed, dtype="float32")
    else:
        shuffle_values = ops.zeros(shape, dtype="float32")

    # When `mask = False`, increase value by 1 so that those indices are placed
    # at the end. Note that `shuffle_values` lies in the range `[0, 1)`, so
    # adding by 1 works out.
    if mask is not None:
        shuffle_values = ops.where(
            mask,
            shuffle_values,
            ops.add(shuffle_values, ops.cast(1, dtype="float32")),
        )

    shuffled_indices = ops.argsort(shuffle_values)
    return shuffled_indices


def sort_by_scores(
    tensors_to_sort: list[types.Tensor],
    scores: types.Tensor,
    mask: Optional[types.Tensor] = None,
    k: Optional[int] = None,
    shuffle_ties: bool = True,
    seed: Optional[Union[int, keras.random.SeedGenerator]] = None,
) -> types.Tensor:
    """
    Utility function for sorting tensors by scores.

    Args:
        tensors_to_sort. list of tensors. All tensors are of shape
            `(batch_size, list_size)`. These tensors are sorted based on
            `scores`.
        scores: tensor. Of shape `(batch_size, list_size)`. The scores to sort
            by.
        k: int. The number of top-ranked items to consider (the 'k' in 'top-k').
            If `None`, `list_size` is used.
        shuffle_ties: bool. Whether to randomly shuffle scores before sorting.
            This is done to break ties.
        seed: int. Seed for shuffling.

    Returns:
        List of sorted tensors (`tensors_to_sort`), sorted using `scores`.
    """
    max_possible_k = ops.shape(scores)[1]
    if k is None:
        k = max_possible_k
    elif isinstance(max_possible_k, int):
        k = min(k, max_possible_k)
    else:
        k = ops.minimum(k, max_possible_k)

    # Shuffle ties randomly, and push masked values to the beginning.
    shuffled_indices = None
    if shuffle_ties or mask is not None:
        shuffled_indices = get_shuffled_indices(
            ops.shape(scores),
            mask=mask,
            shuffle_ties=True,
            seed=seed,
        )
        scores = ops.take_along_axis(scores, shuffled_indices, axis=1)

    # Get top-k indices.
    _, indices = ops.top_k(scores, k=k, sorted=True)

    # If we shuffled our `scores` tensor, we need to get the correct indices
    # by indexing into `shuffled_indices`.
    if shuffled_indices is not None:
        indices = ops.take_along_axis(shuffled_indices, indices, axis=1)

    return [
        ops.take_along_axis(tensor_to_sort, indices, axis=1)
        for tensor_to_sort in tensors_to_sort
    ]


def get_list_weights(
    weights: types.Tensor, relevance: types.Tensor
) -> types.Tensor:
    """Computes per-list weights from provided sample weights.

    Per-list weights are calculated as follows:
    ```
    per_list_weights = sum(weights * relevance) / sum(relevance).
    ```

    For lists where the sum of relevance is 0, a default weight is assigned:
    ```
    sum(per_list_weights) / num(sum(relevance) != 0 AND sum(weights) != 0)
    ```

    If all lists have a sum of relevance equal to 0, the default weight is 1.0.

    As a result of the above computation, this function takes care of the
    following cases:
    - **Uniform Weights:** When all input weights are 1.0, all per-list weights
      will be 1.0, even for lists with no relevant examples. This aligns with
      standard ranking metrics.
    - **Non-zero Weights per List:** If every list has at least one non-zero
      weight, the default weight mechanism is not utilized, which is suitable
      for unbiased metrics.
    - **Mixed Scenarios:** For cases with a mix of lists having zero and
      non-zero relevance and weights, the weights for lists with non-zero
      relevance and weights are proportional to:

      ```
      per_list_weights / sum(per_list_weights) *
      num(sum(relevance) != 0) / num(lists)
      ```

      The rest have weights `1.0 / num(lists)`.

    Args:
        weights: tensor. Weights tensor of shape `(batch_size, list_size)`.
        relevance: tensor. The relevance `Tensor` of shape
            `(batch_size, list_size)`.

    Returns:
        A tensor of shape `(batch_size, 1)`, containing the per-list weights.
    """
    # Calculate if the sum of weights per list is greater than 0.0.
    nonzero_weights = ops.greater(ops.sum(weights, axis=1, keepdims=True), 0.0)
    # Calculate the sum of relevance per list
    per_list_relevance = ops.sum(relevance, axis=1, keepdims=True)
    # Calculate if the sum of relevance per list is greater than 0.0
    nonzero_relevance_condition = ops.greater(per_list_relevance, 0.0)
    # Identify lists where both weights and relevance sums are non-zero.
    nonzero_relevance = ops.cast(
        ops.logical_and(nonzero_weights, nonzero_relevance_condition),
        dtype=weights.dtype,
    )
    # Count the number of lists with non-zero relevance and non-zero weights.
    nonzero_relevance_count = ops.sum(nonzero_relevance, axis=0, keepdims=True)

    # Calculate the per-list weights using the core formula.
    # Numerator: `sum(weights * relevance)` per list
    numerator = ops.sum(ops.multiply(weights, relevance), axis=1, keepdims=True)
    # Denominator: per_list_relevance = sum(relevance) per list
    per_list_weights = ops.divide_no_nan(numerator, per_list_relevance)

    # Calculate the sum of the computed per-list weights.
    sum_weights = ops.sum(per_list_weights, axis=0, keepdims=True)

    # Calculate the average weight to use as default for lists with zero
    # relevance but non-zero weights. If no lists have non-zero relevance,
    # default to 1.0.
    avg_weight = ops.where(
        ops.greater(nonzero_relevance_count, 0.0),
        ops.divide(sum_weights, nonzero_relevance_count),
        ops.cast(1, dtype=sum_weights.dtype),
    )

    # Final assignment of weights based on conditions:
    # 1. If sum(weights) == 0 for a list, the final weight is 0.
    # 2. If sum(weights) > 0 AND sum(relevance) > 0, use the calculated
    # `per_list_weights`.
    # 3. If `sum(weights) > 0` AND `sum(relevance) == 0`, use the calculated
    # `avg_weight`.
    final_weights = ops.where(
        nonzero_weights,
        ops.where(
            nonzero_relevance_condition,
            per_list_weights,
            avg_weight,
        ),
        ops.cast(0, dtype=per_list_weights.dtype),
    )

    return final_weights


@keras.saving.register_keras_serializable()  # type: ignore[misc]
def default_gain_fn(label: types.Tensor) -> types.Tensor:
    return ops.subtract(ops.power(2.0, label), 1.0)


@keras.saving.register_keras_serializable()  # type: ignore[misc]
def default_rank_discount_fn(rank: types.Tensor) -> types.Tensor:
    return ops.divide(
        ops.cast(1, dtype=rank.dtype),
        ops.log2(ops.add(ops.cast(1, dtype=rank.dtype), rank)),
    )


def compute_dcg(
    y_true: types.Tensor,
    sample_weight: types.Tensor,
    gain_fn: Callable[[types.Tensor], types.Tensor] = default_gain_fn,
    rank_discount_fn: Callable[
        [types.Tensor], types.Tensor
    ] = default_rank_discount_fn,
) -> types.Tensor:
    list_size = ops.shape(y_true)[1]
    positions = ops.arange(1, list_size + 1, dtype=y_true.dtype)
    gain = gain_fn(y_true)
    discount = rank_discount_fn(positions)

    return ops.sum(
        ops.multiply(sample_weight, ops.multiply(gain, discount)),
        axis=1,
        keepdims=True,
    )



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from keras_rs.src.losses.pairwise_hinge_loss import (
    PairwiseHingeLoss as PairwiseHingeLoss,
)
from keras_rs.src.losses.pairwise_logistic_loss import (
    PairwiseLogisticLoss as PairwiseLogisticLoss,
)
from keras_rs.src.losses.pairwise_mean_squared_error import (
    PairwiseMeanSquaredError as PairwiseMeanSquaredError,
)
from keras_rs.src.losses.pairwise_soft_zero_one_loss import (
    PairwiseSoftZeroOneLoss as PairwiseSoftZeroOneLoss,
)



from keras import ops

from keras_rs.src import types
from keras_rs.src.api_export import keras_rs_export
from keras_rs.src.losses.pairwise_loss import PairwiseLoss
from keras_rs.src.losses.pairwise_loss import pairwise_loss_subclass_doc_string


@keras_rs_export("keras_rs.losses.PairwiseHingeLoss")
class PairwiseHingeLoss(PairwiseLoss):
    def pairwise_loss(self, pairwise_logits: types.Tensor) -> types.Tensor:
        return ops.relu(ops.subtract(ops.array(1), pairwise_logits))


formula = "loss = sum_{i} sum_{j} I(y_i > y_j) * max(0, 1 - (s_i - s_j))"
explanation = """
    - `max(0, 1 - (s_i - s_j))` is the hinge loss, which penalizes cases where
      the score difference `s_i - s_j` is not sufficiently large when
      `y_i > y_j`.
"""
extra_args = ""
example = """
    With `compile()` API:

    ```python
    model.compile(
        loss=keras_rs.losses.PairwiseHingeLoss(),
        ...
    )
    ```

    As a standalone function with unbatched inputs:

    >>> y_true = np.array([1.0, 0.0, 1.0, 3.0, 2.0])
    >>> y_pred = np.array([1.0, 3.0, 2.0, 4.0, 0.8])
    >>> pairwise_hinge_loss = keras_rs.losses.PairwiseHingeLoss()
    >>> pairwise_hinge_loss(y_true=y_true, y_pred=y_pred)
    2.32000

    With batched inputs using default 'auto'/'sum_over_batch_size' reduction:

    >>> y_true = np.array([[1.0, 0.0, 1.0, 3.0], [0.0, 1.0, 2.0, 3.0]])
    >>> y_pred = np.array([[1.0, 3.0, 2.0, 4.0], [1.0, 1.8, 2.0, 3.0]])
    >>> pairwise_hinge_loss = keras_rs.losses.PairwiseHingeLoss()
    >>> pairwise_hinge_loss(y_true=y_true, y_pred=y_pred)
    0.75

    With masked inputs (useful for ragged inputs):

    >>> y_true = {
    ...     "labels": np.array([[1.0, 0.0, 1.0, 3.0], [0.0, 1.0, 2.0, 3.0]]),
    ...     "mask": np.array(
    ...         [[True, True, True, True], [True, True, False, False]]
    ...     ),
    ... }
    >>> y_pred = np.array([[1.0, 3.0, 2.0, 4.0], [1.0, 1.8, 2.0, 3.0]])
    >>> pairwise_hinge_loss(y_true=y_true, y_pred=y_pred)
    0.64999

    With `sample_weight`:

    >>> y_true = np.array([[1.0, 0.0, 1.0, 3.0], [0.0, 1.0, 2.0, 3.0]])
    >>> y_pred = np.array([[1.0, 3.0, 2.0, 4.0], [1.0, 1.8, 2.0, 3.0]])
    >>> sample_weight = np.array(
    ...     [[2.0, 3.0, 1.0, 1.0], [2.0, 1.0, 0.0, 0.0]]
    ... )
    >>> pairwise_hinge_loss = keras_rs.losses.PairwiseHingeLoss()
    >>> pairwise_hinge_loss(
    ...     y_true=y_true, y_pred=y_pred, sample_weight=sample_weight
    ... )
    1.02499

    Using `'none'` reduction:

    >>> y_true = np.array([[1.0, 0.0, 1.0, 3.0], [0.0, 1.0, 2.0, 3.0]])
    >>> y_pred = np.array([[1.0, 3.0, 2.0, 4.0], [1.0, 1.8, 2.0, 3.0]])
    >>> pairwise_hinge_loss = keras_rs.losses.PairwiseHingeLoss(
    ...     reduction="none"
    ... )
    >>> pairwise_hinge_loss(y_true=y_true, y_pred=y_pred)
    [[3. , 0. , 2. , 0.], [0., 0.20000005, 0.79999995, 0.]]
"""

PairwiseHingeLoss.__doc__ = pairwise_loss_subclass_doc_string.format(
    loss_name="hinge loss",
    formula=formula,
    explanation=explanation,
    extra_args=extra_args,
    example=example,
)



from typing import Any

import keras
from keras import ops

from keras_rs.src import types
from keras_rs.src.api_export import keras_rs_export


@keras_rs_export("keras_rs.layers.SamplingProbabilityCorrection")
class SamplingProbabilityCorrection(keras.layers.Layer):
    """Sampling probability correction.

    Corrects the logits to reflect the sampling probability of negatives.

    Args:
        epsilon: float. Small float added to sampling probability to avoid
            taking the log of zero. Defaults to 1e-6.
        **kwargs: Args to pass to the base class.

    Example:

    ```python
    # Create the layer.
    sampling_probability_correction = (
        keras_rs.layers.SamplingProbabilityCorrection()
    )

    # Correct the logits based on the provided candidate sampling probability.
    logits = sampling_probability_correction(logits, probabilities)
    ```
    """

    def __init__(self, epsilon: float = 1e-6, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.epsilon = epsilon
        self.built = True

    def call(
        self,
        logits: types.Tensor,
        candidate_sampling_probability: types.Tensor,
    ) -> types.Tensor:
        """Corrects input logits to account for candidate sampling probability.

        Args:
            logits: The logits tensor to correct, typically
                `[batch_size, num_candidates]` but can have more dimensions or
                be 1D as `[num_candidates]`.
            candidate_sampling_probability: The sampling probability with the
                same shape as `logits`.

        Returns:
            The corrected logits with the same shape as the input logits.
        """
        return logits - ops.log(
            ops.clip(candidate_sampling_probability, self.epsilon, 1.0)
        )

    def get_config(self) -> dict[str, Any]:
        config: dict[str, Any] = super().get_config()
        config.update({"epsilon": self.epsilon})
        return config



from keras import ops

from keras_rs.src import types
from keras_rs.src.api_export import keras_rs_export
from keras_rs.src.losses.pairwise_loss import PairwiseLoss
from keras_rs.src.losses.pairwise_loss import pairwise_loss_subclass_doc_string


@keras_rs_export("keras_rs.losses.PairwiseSoftZeroOneLoss")
class PairwiseSoftZeroOneLoss(PairwiseLoss):
    def pairwise_loss(self, pairwise_logits: types.Tensor) -> types.Tensor:
        return ops.where(
            ops.greater(pairwise_logits, ops.array(0.0)),
            ops.subtract(ops.array(1.0), ops.sigmoid(pairwise_logits)),
            ops.sigmoid(ops.negative(pairwise_logits)),
        )


formula = "loss = sum_{i} sum_{j} I(y_i > y_j) * (1 - sigmoid(s_i - s_j))"
explanation = """
    - `(1 - sigmoid(s_i - s_j))` represents the soft zero-one loss, which
      approximates the ideal zero-one loss (which would be 1 if `s_i < s_j`
      and 0 otherwise) with a smooth, differentiable function. This makes it
      suitable for gradient-based optimization.
"""
extra_args = ""
example = """
    With `compile()` API:

    ```python
    model.compile(
        loss=keras_rs.losses.PairwiseSoftZeroOneLoss(),
        ...
    )
    ```

    As a standalone function with unbatched inputs:

    >>> y_true = np.array([1.0, 0.0, 1.0, 3.0, 2.0])
    >>> y_pred = np.array([1.0, 3.0, 2.0, 4.0, 0.8])
    >>> pairwise_soft_zero_one_loss = keras_rs.losses.PairwiseSoftZeroOneLoss()
    >>> pairwise_soft_zero_one_loss(y_true=y_true, y_pred=y_pred)
    0.86103

    With batched inputs using default 'auto'/'sum_over_batch_size' reduction:

    >>> y_true = np.array([[1.0, 0.0, 1.0, 3.0], [0.0, 1.0, 2.0, 3.0]])
    >>> y_pred = np.array([[1.0, 3.0, 2.0, 4.0], [1.0, 1.8, 2.0, 3.0]])
    >>> pairwise_soft_zero_one_loss = keras_rs.losses.PairwiseSoftZeroOneLoss()
    >>> pairwise_soft_zero_one_loss(y_true=y_true, y_pred=y_pred)
    0.46202

    With masked inputs (useful for ragged inputs):

    >>> y_true = {
    ...     "labels": np.array([[1.0, 0.0, 1.0, 3.0], [0.0, 1.0, 2.0, 3.0]]),
    ...     "mask": np.array(
    ...         [[True, True, True, True], [True, True, False, False]]
    ...     ),
    ... }
    >>> y_pred = np.array([[1.0, 3.0, 2.0, 4.0], [1.0, 1.8, 2.0, 3.0]])
    >>> pairwise_soft_zero_one_loss(y_true=y_true, y_pred=y_pred)
    0.29468

    With `sample_weight`:

    >>> y_true = np.array([[1.0, 0.0, 1.0, 3.0], [0.0, 1.0, 2.0, 3.0]])
    >>> y_pred = np.array([[1.0, 3.0, 2.0, 4.0], [1.0, 1.8, 2.0, 3.0]])
    >>> sample_weight = np.array(
    ...     [[2.0, 3.0, 1.0, 1.0], [2.0, 1.0, 0.0, 0.0]]
    ... )
    >>> pairwise_soft_zero_one_loss = keras_rs.losses.PairwiseSoftZeroOneLoss()
    >>> pairwise_soft_zero_one_loss(
    ...     y_true=y_true, y_pred=y_pred, sample_weight=sample_weight
    ... )
    0.40478

    Using `'none'` reduction:

    >>> y_true = np.array([[1.0, 0.0, 1.0, 3.0], [0.0, 1.0, 2.0, 3.0]])
    >>> y_pred = np.array([[1.0, 3.0, 2.0, 4.0], [1.0, 1.8, 2.0, 3.0]])
    >>> pairwise_soft_zero_one_loss = keras_rs.losses.PairwiseSoftZeroOneLoss(
    ...     reduction="none"
    ... )
    >>> pairwise_soft_zero_one_loss(y_true=y_true, y_pred=y_pred)
    [
        [0.8807971 , 0., 0.73105854, 0.43557024],
        [0., 0.31002545, 0.7191075 , 0.61961967]
    ]
"""

PairwiseSoftZeroOneLoss.__doc__ = pairwise_loss_subclass_doc_string.format(
    loss_name="soft zero-one loss",
    formula=formula,
    explanation=explanation,
    extra_args=extra_args,
    example=example,
)



from typing import Any, Optional, Text, Union

import keras
from keras import ops

from keras_rs.src import types
from keras_rs.src.api_export import keras_rs_export
from keras_rs.src.utils.keras_utils import clone_initializer


@keras_rs_export("keras_rs.layers.FeatureCross")
class FeatureCross(keras.layers.Layer):
    """FeatureCross layer in Deep & Cross Network (DCN).

    A layer that creates explicit and bounded-degree feature interactions
    efficiently. The `call` method accepts two inputs: `x0` contains the
    original features; the second input `xi` is the output of the previous
    `FeatureCross` layer in the stack, i.e., the i-th `FeatureCross` layer.
    For the first `FeatureCross` layer in the stack, `x0 = xi`.

    The output is `x_{i+1} = x0 .* (W * x_i + bias + diag_scale * x_i) + x_i`,
    where .* denotes element-wise multiplication. W could be a full-rank
    matrix, or a low-rank matrix `U*V` to reduce the computational cost, and
    `diag_scale` increases the diagonal of W to improve training stability (
    especially for the low-rank case).

    Args:
        projection_dim: int. Dimension for down-projecting the input to reduce
            computational cost. If `None` (default), the full matrix, `W`
            (with shape `(input_dim, input_dim)`) is used. Otherwise, a low-rank
            matrix `W = U*V` will be used, where `U` is of shape
            `(input_dim, projection_dim)` and `V` is of shape
            `(projection_dim, input_dim)`. `projection_dim` need to be smaller
            than `input_dim//2` to improve the model efficiency. In practice,
            we've observed that `projection_dim = input_dim//4` consistently
            preserved the accuracy of a full-rank version.
        diag_scale: non-negative float. Used to increase the diagonal of the
            kernel W by `diag_scale`, i.e., `W + diag_scale * I`, where I is the
            identity matrix. Defaults to `None`.
        use_bias: bool. Whether to add a bias term for this layer. Defaults to
            `True`.
        pre_activation: string or `keras.activations`. Activation applied to
            output matrix of the layer, before multiplication with the input.
            Can be used to control the scale of the layer's outputs and
            improve stability. Defaults to `None`.
        kernel_initializer: string or `keras.initializers` initializer.
            Initializer to use for the kernel matrix. Defaults to
            `"glorot_uniform"`.
        bias_initializer: string or `keras.initializers` initializer.
            Initializer to use for the bias vector. Defaults to `"ones"`.
        kernel_regularizer: string or `keras.regularizer` regularizer.
            Regularizer to use for the kernel matrix.
        bias_regularizer: string or `keras.regularizer` regularizer.
            Regularizer to use for the bias vector.
        **kwargs: Args to pass to the base class.

    Example:

    ```python
    # 1. Simple forward pass
    batch_size = 2
    embedding_dim = 32
    feature1 = np.random.randn(batch_size, embedding_dim)
    feature2 = np.random.randn(batch_size, embedding_dim)
    crossed_features = keras_rs.layers.FeatureCross()(feature1, feature2)

    # 2. After embedding layer in a model
    vocabulary_size = 32
    embedding_dim = 6

    # Create a simple model containing the layer.
    inputs = keras.Input(shape=(), name='indices', dtype="int32")
    x0 = keras.layers.Embedding(
        input_dim=vocabulary_size,
        output_dim=embedding_dim
    )(inputs)
    x1 = keras_rs.layers.FeatureCross()(x0, x0)
    x2 = keras_rs.layers.FeatureCross()(x0, x1)
    logits = keras.layers.Dense(units=10)(x2)
    model = keras.Model(inputs, logits)

    # Call the model on the inputs.
    batch_size = 2
    input_data = np.random.randint(0, vocabulary_size, size=(batch_size,))
    outputs = model(input_data)
    ```

    References:
    - [R. Wang et al.](https://arxiv.org/abs/2008.13535)
    - [R. Wang et al.](https://arxiv.org/abs/1708.05123)
    """

    def __init__(
        self,
        projection_dim: Optional[int] = None,
        diag_scale: Optional[float] = 0.0,
        use_bias: bool = True,
        pre_activation: Optional[Union[str, keras.layers.Activation]] = None,
        kernel_initializer: Union[
            Text, keras.initializers.Initializer
        ] = "glorot_uniform",
        bias_initializer: Union[Text, keras.initializers.Initializer] = "zeros",
        kernel_regularizer: Union[
            Text, None, keras.regularizers.Regularizer
        ] = None,
        bias_regularizer: Union[
            Text, None, keras.regularizers.Regularizer
        ] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)

        # Passed args.
        self.projection_dim = projection_dim
        self.diag_scale = diag_scale
        self.use_bias = use_bias
        self.pre_activation = keras.activations.get(pre_activation)
        self.kernel_initializer = keras.initializers.get(kernel_initializer)
        self.bias_initializer = keras.initializers.get(bias_initializer)
        self.kernel_regularizer = keras.regularizers.get(kernel_regularizer)
        self.bias_regularizer = keras.regularizers.get(bias_regularizer)

        # Other args.
        self.supports_masking = True

        if self.diag_scale is not None and self.diag_scale < 0.0:
            raise ValueError(
                "`diag_scale` should be non-negative. Received: "
                f"`diag_scale={self.diag_scale}`"
            )

    def build(self, input_shape: types.TensorShape) -> None:
        last_dim = input_shape[-1]

        if self.projection_dim is not None:
            self.down_proj_dense = keras.layers.Dense(
                units=self.projection_dim,
                use_bias=False,
                kernel_initializer=clone_initializer(self.kernel_initializer),
                kernel_regularizer=self.kernel_regularizer,
                dtype=self.dtype_policy,
            )

        self.dense = keras.layers.Dense(
            units=last_dim,
            activation=self.pre_activation,
            use_bias=self.use_bias,
            kernel_initializer=clone_initializer(self.kernel_initializer),
            bias_initializer=clone_initializer(self.bias_initializer),
            kernel_regularizer=self.kernel_regularizer,
            bias_regularizer=self.bias_regularizer,
            dtype=self.dtype_policy,
        )

        self.built = True

    def call(
        self, x0: types.Tensor, x: Optional[types.Tensor] = None
    ) -> types.Tensor:
        """Forward pass of the cross layer.

        Args:
            x0: a Tensor. The input to the cross layer. N-rank tensor
                with shape `(batch_size, ..., input_dim)`.
            x: a Tensor. Optional. If provided, the layer will compute
                crosses between x0 and x. Otherwise, the layer will
                compute crosses between x0 and itself. Should have the same
                shape as `x0`.

        Returns:
            Tensor of crosses, with the same shape as `x0`.
        """

        if x is None:
            x = x0

        if x0.shape != x.shape:
            raise ValueError(
                "`x0` and `x` should have the same shape. Received: "
                f"`x.shape` = {x.shape}, `x0.shape` = {x0.shape}"
            )

        # Project to a lower dimension.
        if self.projection_dim is None:
            output = x
        else:
            output = self.down_proj_dense(x)

        output = self.dense(output)

        output = ops.cast(output, self.compute_dtype)

        if self.diag_scale:
            output = ops.add(output, ops.multiply(self.diag_scale, x))

        return ops.add(ops.multiply(x0, output), x)

    def get_config(self) -> dict[str, Any]:
        config: dict[str, Any] = super().get_config()

        config.update(
            {
                "projection_dim": self.projection_dim,
                "diag_scale": self.diag_scale,
                "use_bias": self.use_bias,
                "pre_activation": keras.activations.serialize(
                    self.pre_activation
                ),
                "kernel_initializer": keras.initializers.serialize(
                    self.kernel_initializer
                ),
                "bias_initializer": keras.initializers.serialize(
                    self.bias_initializer
                ),
                "kernel_regularizer": keras.regularizers.serialize(
                    self.kernel_regularizer
                ),
                "bias_regularizer": keras.regularizers.serialize(
                    self.bias_regularizer
                ),
            }
        )

        return config



from keras import ops

from keras_rs.src import types
from keras_rs.src.api_export import keras_rs_export
from keras_rs.src.losses.pairwise_loss import PairwiseLoss
from keras_rs.src.losses.pairwise_loss import pairwise_loss_subclass_doc_string


@keras_rs_export("keras_rs.losses.PairwiseLogisticLoss")
class PairwiseLogisticLoss(PairwiseLoss):
    def pairwise_loss(self, pairwise_logits: types.Tensor) -> types.Tensor:
        return ops.add(
            ops.relu(ops.negative(pairwise_logits)),
            ops.log(
                ops.add(
                    ops.array(1),
                    ops.exp(ops.negative(ops.abs(pairwise_logits))),
                )
            ),
        )


formula = "loss = sum_{i} sum_{j} I(y_i > y_j) * log(1 + exp(-(s_i - s_j)))"
explanation = """
    - `log(1 + exp(-(s_i - s_j)))` is the logistic loss, which penalizes
      cases where the score difference `s_i - s_j` is not sufficiently large
      when `y_i > y_j`. This function provides a smooth approximation of the
      ideal step function, making it suitable for gradient-based optimization.
"""
extra_args = ""
example = """
    With `compile()` API:

    ```python
    model.compile(
        loss=keras_rs.losses.PairwiseLogisticLoss(),
        ...
    )
    ```

    As a standalone function with unbatched inputs:

    >>> y_true = np.array([1.0, 0.0, 1.0, 3.0, 2.0])
    >>> y_pred = np.array([1.0, 3.0, 2.0, 4.0, 0.8])
    >>> pairwise_logistic_loss = keras_rs.losses.PairwiseLogisticLoss()
    >>> pairwise_logistic_loss(y_true=y_true, y_pred=y_pred)
    >>> 1.70708

    With batched inputs using default 'auto'/'sum_over_batch_size' reduction:

    >>> y_true = np.array([[1.0, 0.0, 1.0, 3.0], [0.0, 1.0, 2.0, 3.0]])
    >>> y_pred = np.array([[1.0, 3.0, 2.0, 4.0], [1.0, 1.8, 2.0, 3.0]])
    >>> pairwise_logistic_loss = keras_rs.losses.PairwiseLogisticLoss()
    >>> pairwise_logistic_loss(y_true=y_true, y_pred=y_pred)
    0.73936

    With masked inputs (useful for ragged inputs):

    >>> y_true = {
    ...     "labels": np.array([[1.0, 0.0, 1.0, 3.0], [0.0, 1.0, 2.0, 3.0]]),
    ...     "mask": np.array(
    ...         [[True, True, True, True], [True, True, False, False]]
    ...     ),
    ... }
    >>> y_pred = np.array([[1.0, 3.0, 2.0, 4.0], [1.0, 1.8, 2.0, 3.0]])
    >>> pairwise_logistic_loss(y_true=y_true, y_pred=y_pred)
    0.53751

    With `sample_weight`:

    >>> y_true = np.array([[1.0, 0.0, 1.0, 3.0], [0.0, 1.0, 2.0, 3.0]])
    >>> y_pred = np.array([[1.0, 3.0, 2.0, 4.0], [1.0, 1.8, 2.0, 3.0]])
    >>> sample_weight = np.array(
    ...     [[2.0, 3.0, 1.0, 1.0], [2.0, 1.0, 0.0, 0.0]]
    ... )
    >>> pairwise_logistic_loss = keras_rs.losses.PairwiseLogisticLoss()
    >>> pairwise_logistic_loss(
    ...     y_true=y_true, y_pred=y_pred, sample_weight=sample_weight
    ... )
    >>> 0.80337

    Using `'none'` reduction:

    >>> y_true = np.array([[1.0, 0.0, 1.0, 3.0], [0.0, 1.0, 2.0, 3.0]])
    >>> y_pred = np.array([[1.0, 3.0, 2.0, 4.0], [1.0, 1.8, 2.0, 3.0]])
    >>> pairwise_logistic_loss = keras_rs.losses.PairwiseLogisticLoss(
    ...     reduction="none"
    ... )
    >>> pairwise_logistic_loss(y_true=y_true, y_pred=y_pred)
    [[2.126928, 0., 1.3132616, 0.48877698], [0., 0.20000005, 0.79999995, 0.]]
"""

PairwiseLogisticLoss.__doc__ = pairwise_loss_subclass_doc_string.format(
    loss_name="logistic loss",
    formula=formula,
    explanation=explanation,
    extra_args=extra_args,
    example=example,
)



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from keras_rs.src.layers.feature_interaction.dot_interaction import (
    DotInteraction as DotInteraction,
)
from keras_rs.src.layers.feature_interaction.feature_cross import (
    FeatureCross as FeatureCross,
)
from keras_rs.src.layers.retrieval.brute_force_retrieval import (
    BruteForceRetrieval as BruteForceRetrieval,
)
from keras_rs.src.layers.retrieval.hard_negative_mining import (
    HardNegativeMining as HardNegativeMining,
)
from keras_rs.src.layers.retrieval.remove_accidental_hits import (
    RemoveAccidentalHits as RemoveAccidentalHits,
)
from keras_rs.src.layers.retrieval.retrieval import Retrieval as Retrieval
from keras_rs.src.layers.retrieval.sampling_probability_correction import (
    SamplingProbabilityCorrection as SamplingProbabilityCorrection,
)



from keras_rs.src.api_export import keras_rs_export

# Unique source of truth for the version number.
__version__ = "0.1.0"


@keras_rs_export("keras_rs.version")
def version() -> str:
    return __version__



from typing import Any

import keras
from keras import ops

from keras_rs.src import types
from keras_rs.src.api_export import keras_rs_export
from keras_rs.src.utils.keras_utils import check_shapes_compatible


@keras_rs_export("keras_rs.layers.DotInteraction")
class DotInteraction(keras.layers.Layer):
    """Dot interaction layer present in DLRM.

    This layer computes distinct dot product ("feature interactions") for every
    pair of features. If `self_interaction` is True, we calculate dot products
    of the form `dot(e_i, e_j)` for `i <= j`, and `dot(e_i, e_j)` for `i < j`,
    otherwise. `e_i` denotes representation of feature `i`. The layer can be
    used to build the DLRM model.

    Args:
        self_interaction: bool. Indicates whether features should
            "self-interact". If it is True, then the diagonal entries of the
            interaction matrix are also taken.
        skip_gather: bool. If it's set, then the upper triangular part of the
            interaction matrix is set to 0. The output will be of shape
            `[batch_size, num_features * num_features]` of which half of the
            entries will be zeros. Otherwise, the output will be only the lower
            triangular part of the interaction matrix. The latter saves space
            but is much slower.
        **kwargs: Args to pass to the base class.

    Example:

    ```python
    # 1. Simple forward pass
    batch_size = 2
    embedding_dim = 32
    feature1 = np.random.randn(batch_size, embedding_dim)
    feature2 = np.random.randn(batch_size, embedding_dim)
    feature3 = np.random.randn(batch_size, embedding_dim)
    feature_interactions = keras_rs.layers.DotInteraction()(
        [feature1, feature2, feature3]
    )

    # 2. After embedding layer in a model
    vocabulary_size = 32
    embedding_dim = 6

    # Create a simple model containing the layer.
    feature_input_1 = keras.Input(shape=(), name='indices_1', dtype="int32")
    feature_input_2 = keras.Input(shape=(), name='indices_2', dtype="int32")
    feature_input_3 = keras.Input(shape=(), name='indices_3', dtype="int32")
    x1 = keras.layers.Embedding(
        input_dim=vocabulary_size,
        output_dim=embedding_dim
    )(feature_input_1)
    x2 = keras.layers.Embedding(
        input_dim=vocabulary_size,
        output_dim=embedding_dim
    )(feature_input_2)
    x3 = keras.layers.Embedding(
        input_dim=vocabulary_size,
        output_dim=embedding_dim
    )(feature_input_3)
    feature_interactions = keras_rs.layers.DotInteraction()([x1, x2, x3])
    output = keras.layers.Dense(units=10)(x2)
    model = keras.Model(
        [feature_input_1, feature_input_2, feature_input_3], output
    )

    # Call the model on the inputs.
    batch_size = 2
    f1 = np.random.randint(0, vocabulary_size, size=(batch_size,))
    f2 = np.random.randint(0, vocabulary_size, size=(batch_size,))
    f3 = np.random.randint(0, vocabulary_size, size=(batch_size,))
    outputs = model([f1, f2, f3])
    ```

    References:
    - [M. Naumov et al.](https://arxiv.org/abs/1906.00091)
    """

    def __init__(
        self,
        self_interaction: bool = False,
        skip_gather: bool = False,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)

        # Attributes.
        self.self_interaction = self_interaction
        self.skip_gather = skip_gather

    def _generate_tril_mask(
        self, pairwise_interaction_matrix: types.Tensor
    ) -> types.Tensor:
        """Generates lower triangular mask."""

        # If `self.self_interaction` is `True`, keep the main diagonal.
        k = -1
        if self.self_interaction:
            k = 0

        # Typecast k from Python int to tensor, because `ops.tril` uses
        # `tf.cond` (which requires tensors).
        # TODO (abheesht): Remove typecast once fix is merged in core Keras.
        if keras.config.backend() == "tensorflow":
            k = ops.array(k)
        tril_mask = ops.tril(
            ops.ones_like(pairwise_interaction_matrix, dtype=bool),
            k=k,
        )

        return tril_mask

    def _get_lower_triangular_indices(self, num_features: int) -> list[int]:
        """Python function which generates indices to get the lower triangular
        matrix as if it were flattened.
        """
        flattened_indices = []
        for i in range(num_features):
            k = i
            # if `self.self_interaction` is `True`, keep the main diagonal.
            if self.self_interaction:
                k += 1
            for j in range(k):
                flattened_index = i * num_features + j
                flattened_indices.append(flattened_index)

        return flattened_indices

    def call(self, inputs: list[types.Tensor]) -> types.Tensor:
        """Forward pass of the dot interaction layer.

        Args:
            inputs: list. Every element in the list represents a feature tensor
                of shape `[batch_size, feature_dim]`. All tensors in the list
                must have the same shape.

        Returns:
            Tensor representing feature interactions. The shape of the tensor is
            `[batch_size, k]` where `k` is
            `num_features * num_features` if `skip_gather` is `True`. Otherwise,
            `k` is `num_features * (num_features + 1) / 2` if
            `self_interaction` is `True`, and
            `num_features * (num_features - 1) / 2` if not.
        """

        # Check if all feature tensors have the same shape and are of rank 2.
        shape = ops.shape(inputs[0])
        for idx, tensor in enumerate(inputs):
            other_shape = ops.shape(tensor)

            if len(shape) != 2:
                raise ValueError(
                    "All feature tensors inside `inputs` should have rank 2. "
                    f"Received rank {len(shape)} at index {idx}."
                )

            if not check_shapes_compatible(shape, other_shape):
                raise ValueError(
                    "All feature tensors in `inputs` should have the same "
                    f"shape. Found at least one conflict: shape = {shape} at "
                    f"index 0 and shape = {ops.shape(tensor)} at index {idx}."
                )

        # `(batch_size, num_features, feature_dim)`
        features = ops.stack(inputs, axis=1)

        batch_size, num_features, _ = ops.shape(features)

        # Compute the dot product to get feature interactions. The shape here is
        # `(batch_size, num_features, num_features)`.
        pairwise_interaction_matrix = ops.matmul(
            features, ops.transpose(features, axes=(0, 2, 1))
        )

        # Set the upper triangle entries to 0, if `self.skip_gather` is True.
        # Else, "pick" only the lower triangle entries.
        if self.skip_gather:
            tril_mask = self._generate_tril_mask(pairwise_interaction_matrix)

            activations = ops.multiply(
                pairwise_interaction_matrix,
                ops.cast(tril_mask, dtype=pairwise_interaction_matrix.dtype),
            )
            # Rank-2 tensor.
            activations = ops.reshape(
                activations, (batch_size, num_features * num_features)
            )
        else:
            flattened_indices = self._get_lower_triangular_indices(num_features)
            pairwise_interaction_matrix_flattened = ops.reshape(
                pairwise_interaction_matrix,
                (batch_size, num_features * num_features),
            )
            activations = ops.take(
                pairwise_interaction_matrix_flattened,
                flattened_indices,
                axis=-1,
            )

        return activations

    def compute_output_shape(
        self, input_shape: list[types.TensorShape]
    ) -> types.TensorShape:
        num_features = len(input_shape)
        batch_size = input_shape[0][0]

        # Determine the number of pairwise interactions
        if self.self_interaction:
            output_dim = num_features * (num_features + 1) // 2
        else:
            output_dim = num_features * (num_features - 1) // 2

        if self.skip_gather:
            output_dim = num_features * num_features

        return (batch_size, output_dim)

    def get_config(self) -> dict[str, Any]:
        config: dict[str, Any] = super().get_config()

        config.update(
            {
                "self_interaction": self.self_interaction,
                "skip_gather": self.skip_gather,
            }
        )

        return config



from typing import Any

import keras
import ml_dtypes
from keras import ops

from keras_rs.src import types
from keras_rs.src.api_export import keras_rs_export

MAX_FLOAT = ml_dtypes.finfo("float32").max / 100.0


@keras_rs_export("keras_rs.layers.HardNegativeMining")
class HardNegativeMining(keras.layers.Layer):
    """Filter logits and labels to return hard negatives.

    The output will include logits and labels for the requested number of hard
    negatives as well as the positive candidate.

    Args:
        num_hard_negatives: How many hard negatives to return.
        **kwargs: Args to pass to the base class.

    Example:

    ```python
    # Create layer with the configured number of hard negatives to mine.
    hard_negative_mining = keras_rs.layers.HardNegativeMining(
        num_hard_negatives=10
    )

    # This will retrieve the top 10 negative candidates plus the positive
    # candidate from `labels` for each row.
    out_logits, out_labels = hard_negative_mining(in_logits, in_labels)
    ```
    """

    def __init__(self, num_hard_negatives: int, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._num_hard_negatives = num_hard_negatives
        self.built = True

    def call(
        self, logits: types.Tensor, labels: types.Tensor
    ) -> tuple[types.Tensor, types.Tensor]:
        """Filters logits and labels with per-query hard negative mining.

        The result will include logits and labels for `num_hard_negatives`
        negatives as well as the positive candidate.

        Args:
            logits: The logits tensor, typically `[batch_size, num_candidates]`
                but can have more dimensions or be 1D as `[num_candidates]`.
            labels: The one-hot labels tensor, must be the same shape as
                `logits`.

        Returns:
            A tuple containing two tensors with the last dimension of
            `num_candidates` replaced with `num_hard_negatives + 1`.

        * logits: `[..., num_hard_negatives + 1]` tensor of logits.
        * labels: `[..., num_hard_negatives + 1]` one-hot tensor of labels.
        """

        # Number of sampled logits, i.e, the number of hard negatives to be
        # sampled (k) + number of true logit (1) per query, capped by batch
        # size.
        num_logits = ops.shape(logits)[-1]
        if isinstance(num_logits, int):
            num_sampled = min(self._num_hard_negatives + 1, num_logits)
        else:
            num_sampled = ops.minimum(self._num_hard_negatives + 1, num_logits)
        # To gather indices of top k negative logits per row (query) in logits,
        # true logits need to be excluded. First replace the true logits
        # (corresponding to positive labels) with a large score value and then
        # select the top k + 1 logits from each row so that selected indices
        # include the indices of true logit + top k negative logits. This
        # approach is to avoid using inefficient masking when excluding true
        # logits.

        # For each query, get the indices of the logits which have the highest
        # k + 1 logit values, including the highest k negative logits and one
        # true logit.
        _, indices = ops.top_k(
            ops.add(logits, ops.multiply(labels, MAX_FLOAT)),
            k=num_sampled,
            sorted=False,
        )

        # Gather sampled logits and corresponding labels.
        logits = ops.take_along_axis(logits, indices, axis=-1)
        labels = ops.take_along_axis(labels, indices, axis=-1)

        return logits, labels



from typing import Any, Callable, Optional

from keras import ops
from keras.saving import deserialize_keras_object
from keras.saving import serialize_keras_object

from keras_rs.src import types
from keras_rs.src.api_export import keras_rs_export
from keras_rs.src.metrics.ranking_metric import RankingMetric
from keras_rs.src.metrics.ranking_metric import (
    ranking_metric_subclass_doc_string,
)
from keras_rs.src.metrics.ranking_metric import (
    ranking_metric_subclass_doc_string_post_desc,
)
from keras_rs.src.metrics.ranking_metrics_utils import compute_dcg
from keras_rs.src.metrics.ranking_metrics_utils import default_gain_fn
from keras_rs.src.metrics.ranking_metrics_utils import default_rank_discount_fn
from keras_rs.src.metrics.ranking_metrics_utils import get_list_weights
from keras_rs.src.metrics.ranking_metrics_utils import sort_by_scores
from keras_rs.src.utils.doc_string_utils import format_docstring


@keras_rs_export("keras_rs.metrics.NDCG")
class NDCG(RankingMetric):
    def __init__(
        self,
        k: Optional[int] = None,
        gain_fn: Callable[[types.Tensor], types.Tensor] = default_gain_fn,
        rank_discount_fn: Callable[
            [types.Tensor], types.Tensor
        ] = default_rank_discount_fn,
        **kwargs: Any,
    ) -> None:
        super().__init__(k=k, **kwargs)

        self.gain_fn = gain_fn
        self.rank_discount_fn = rank_discount_fn

    def compute_metric(
        self,
        y_true: types.Tensor,
        y_pred: types.Tensor,
        mask: types.Tensor,
        sample_weight: types.Tensor,
    ) -> types.Tensor:
        sorted_y_true, sorted_weights = sort_by_scores(
            tensors_to_sort=[y_true, sample_weight],
            scores=y_pred,
            k=self.k,
            mask=mask,
            shuffle_ties=self.shuffle_ties,
            seed=self.seed_generator,
        )

        dcg = compute_dcg(
            y_true=sorted_y_true,
            sample_weight=sorted_weights,
            gain_fn=self.gain_fn,
            rank_discount_fn=self.rank_discount_fn,
        )

        weighted_gains = ops.multiply(
            sample_weight,
            self.gain_fn(y_true),
        )
        ideal_sorted_y_true, ideal_sorted_weights = sort_by_scores(
            tensors_to_sort=[y_true, sample_weight],
            scores=weighted_gains,
            k=self.k,
            mask=mask,
            shuffle_ties=self.shuffle_ties,
            seed=self.seed_generator,
        )
        ideal_dcg = compute_dcg(
            y_true=ideal_sorted_y_true,
            sample_weight=ideal_sorted_weights,
            gain_fn=self.gain_fn,
            rank_discount_fn=self.rank_discount_fn,
        )
        per_list_ndcg = ops.divide_no_nan(dcg, ideal_dcg)

        per_list_weights = get_list_weights(
            weights=sample_weight, relevance=self.gain_fn(y_true)
        )

        return per_list_ndcg, per_list_weights

    def get_config(self) -> dict[str, Any]:
        config: dict[str, Any] = super().get_config()
        config.update(
            {
                "gain_fn": serialize_keras_object(self.gain_fn),
                "rank_discount_fn": serialize_keras_object(
                    self.rank_discount_fn
                ),
            }
        )
        return config

    @classmethod
    def from_config(cls, config: dict[str, Any]) -> "NDCG":
        config["gain_fn"] = deserialize_keras_object(config["gain_fn"])
        config["rank_discount_fn"] = deserialize_keras_object(
            config["rank_discount_fn"]
        )
        return cls(**config)


concept_sentence = (
    "It normalizes the Discounted Cumulative Gain (DCG) with the Ideal "
    "Discounted Cumulative Gain (IDCG) for each list"
)
relevance_type = (
    "graded relevance scores (non-negative numbers where higher values "
    "indicate greater relevance)"
)
score_range_interpretation = (
    "A normalized score (between 0 and 1) is returned. A score of 1 "
    "represents the perfect ranking according to true relevance (within the "
    "top-k), while 0 typically represents a ranking with no relevant items. "
    "Higher scores indicate better ranking quality relative to the best "
    "possible ranking"
)

formula = """
```
nDCG@k = DCG@k / IDCG@k
```

where DCG@k is calculated based on the predicted ranking (`y_pred`):

```
DCG@k(y') = sum_{i=1}^{k} (gain_fn(y'_i) / rank_discount_fn(i))
```

And IDCG@k is the Ideal DCG, calculated using the same formula but on items
sorted perfectly by their *true relevance* (`y_true`):

```
IDCG@k(y'') = sum_{i=1}^{k} (gain_fn(y''_i) / rank_discount_fn(i))
```

where:
- `y'_i`: True relevance of the item at rank `i` in the ranking induced by
  `y_pred`.
- `y''_i` True relevance of the item at rank `i` in the *ideal* ranking (sorted
  by `y_true` descending).
- `gain_fn` is the user-provided function mapping relevance to gain. The default
  function (`default_gain_fn`) is typically equivalent to `lambda y: 2**y - 1`.
- `rank_discount_fn` is the user-provided function mapping rank `i` (1-based) to
  a discount value. The default function (`default_rank_discount_fn`) is
  typically equivalent to `lambda rank: 1 / log2(rank + 1)`.
- If IDCG@k is 0 (e.g., no relevant items), nDCG@k is defined as 0.
- The final result often aggregates these per-list nDCG scores, potentially
  involving normalization by list-specific weights, to produce a weighted
  average."""
extra_args = """
        gain_fn: callable. Maps relevance scores (`y_true`) to gain values. The
            default implements `2**y - 1`.
        rank_discount_fn: function. Maps rank positions to discount
            values. The default (`default_rank_discount_fn`) implements
            `1 / log2(rank + 1)`."""
example = """
    >>> batch_size = 2
    >>> list_size = 5
    >>> labels = np.random.randint(0, 3, size=(batch_size, list_size))
    >>> scores = np.random.random(size=(batch_size, list_size))
    >>> metric = keras_rs.metrics.NDCG()(
    ...     y_true=labels, y_pred=scores
    ... )

    Mask certain elements (can be used for uneven inputs):

    >>> batch_size = 2
    >>> list_size = 5
    >>> labels = np.random.randint(0, 3, size=(batch_size, list_size))
    >>> scores = np.random.random(size=(batch_size, list_size))
    >>> mask = np.random.randint(0, 2, size=(batch_size, list_size), dtype=bool)
    >>> metric = keras_rs.metrics.NDCG()(
    ...     y_true={"labels": labels, "mask": mask}, y_pred=scores
    ... )
"""

NDCG.__doc__ = format_docstring(
    ranking_metric_subclass_doc_string,
    width=80,
    metric_name="Normalised Discounted Cumulative Gain",
    metric_abbreviation="nDCG",
    concept_sentence=concept_sentence,
    relevance_type=relevance_type,
    score_range_interpretation=score_range_interpretation,
    formula=formula,
    extra_args=extra_args,
) + ranking_metric_subclass_doc_string_post_desc.format(
    extra_args=extra_args, example=example
)



from keras import ops

from keras_rs.src import types
from keras_rs.src.api_export import keras_rs_export
from keras_rs.src.metrics.ranking_metric import RankingMetric
from keras_rs.src.metrics.ranking_metric import (
    ranking_metric_subclass_doc_string,
)
from keras_rs.src.metrics.ranking_metric import (
    ranking_metric_subclass_doc_string_post_desc,
)
from keras_rs.src.metrics.ranking_metrics_utils import get_list_weights
from keras_rs.src.metrics.ranking_metrics_utils import sort_by_scores
from keras_rs.src.utils.doc_string_utils import format_docstring


@keras_rs_export("keras_rs.metrics.MeanAveragePrecision")
class MeanAveragePrecision(RankingMetric):
    def compute_metric(
        self,
        y_true: types.Tensor,
        y_pred: types.Tensor,
        mask: types.Tensor,
        sample_weight: types.Tensor,
    ) -> types.Tensor:
        relevance = ops.cast(
            ops.greater_equal(y_true, ops.cast(1, dtype=y_true.dtype)),
            dtype=y_pred.dtype,
        )
        sorted_relevance, sorted_weights = sort_by_scores(
            tensors_to_sort=[relevance, sample_weight],
            scores=y_pred,
            mask=mask,
            k=self.k,
            shuffle_ties=self.shuffle_ties,
            seed=self.seed_generator,
        )
        per_list_relevant_counts = ops.cumsum(sorted_relevance, axis=1)
        per_list_cutoffs = ops.cumsum(ops.ones_like(sorted_relevance), axis=1)
        per_list_precisions = ops.divide_no_nan(
            per_list_relevant_counts, per_list_cutoffs
        )

        total_precision = ops.sum(
            ops.multiply(
                per_list_precisions,
                ops.multiply(sorted_weights, sorted_relevance),
            ),
            axis=1,
            keepdims=True,
        )

        # Compute the total relevance.
        total_relevance = ops.sum(
            ops.multiply(sample_weight, relevance), axis=1, keepdims=True
        )

        per_list_map = ops.divide_no_nan(total_precision, total_relevance)

        per_list_weights = get_list_weights(sample_weight, relevance)

        return per_list_map, per_list_weights


concept_sentence = (
    "It calculates the average of precision values computed after each "
    "relevant item present in the ranked list"
)
relevance_type = "binary indicators (0 or 1) of relevance"
score_range_interpretation = (
    "Scores range from 0 to 1, with higher values indicating that relevant "
    "items are generally positioned higher in the ranking"
)

formula = """
The formula for average precision is defined below. MAP is the mean over average
precision computed for each list.

```
AP(y, s) = sum_j (P@j(y, s) * rel(j)) / sum_i y_i
rel(j) = y_i if rank(s_i) = j
```

where:
- `j` represents the rank position (starting from 1).
- `sum_j` indicates a summation over all ranks `j` from 1 up to the list
  size (or `k`).
- `P@j(y, s)` denotes the Precision at rank `j`, calculated as the
  number of relevant items found within the top `j` positions divided by `j`.
- `rel(j)` represents the relevance of the item specifically at rank
  `j`. `rel(j)` is 1 if the item at rank `j` is relevant, and 0 otherwise.
- `y_i` is the true relevance label of the original item `i` before ranking.
- `rank(s_i)` is the rank position assigned to item `i` based on its score
  `s_i`.
- `sum_i y_i` calculates the total number of relevant items in the original
  list `y`."""
extra_args = ""
example = """
    >>> batch_size = 2
    >>> list_size = 5
    >>> labels = np.random.randint(0, 2, size=(batch_size, list_size))
    >>> scores = np.random.random(size=(batch_size, list_size))
    >>> metric = keras_rs.metrics.MeanAveragePrecision()(
    ...     y_true=labels, y_pred=scores
    ... )

    Mask certain elements (can be used for uneven inputs):

    >>> batch_size = 2
    >>> list_size = 5
    >>> labels = np.random.randint(0, 2, size=(batch_size, list_size))
    >>> scores = np.random.random(size=(batch_size, list_size))
    >>> mask = np.random.randint(0, 2, size=(batch_size, list_size), dtype=bool)
    >>> metric = keras_rs.metrics.MeanAveragePrecision()(
    ...     y_true={"labels": labels, "mask": mask}, y_pred=scores
    ... )
"""

MeanAveragePrecision.__doc__ = format_docstring(
    ranking_metric_subclass_doc_string,
    width=80,
    metric_name="Mean Average Precision",
    metric_abbreviation="MAP",
    concept_sentence=concept_sentence,
    relevance_type=relevance_type,
    score_range_interpretation=score_range_interpretation,
    formula=formula,
) + ranking_metric_subclass_doc_string_post_desc.format(
    extra_args=extra_args, example=example
)



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from keras_rs import layers as layers
from keras_rs import losses as losses
from keras_rs import metrics as metrics
from keras_rs.src.version import __version__ as __version__
from keras_rs.src.version import version as version



from typing import Any, Optional, Union

import keras

from keras_rs.src import types
from keras_rs.src.api_export import keras_rs_export
from keras_rs.src.layers.retrieval.retrieval import Retrieval


@keras_rs_export("keras_rs.layers.BruteForceRetrieval")
class BruteForceRetrieval(Retrieval):
    """Brute force top-k retrieval.

    This layer maintains a set of candidates and is able to exactly retrieve the
    top-k candidates for a given query. It does this by computing the scores for
    all of the candidates for the query and extracting the top ones. The
    returned top-k candidates are sorted by score.

    By default, this layer returns a tuple with the top scores and top
    identifiers, but it can be configured to return a single tensor with the top
    identifiers.

    The identifiers for the candidates can be specified as a tensor. If not
    provided, the IDs used are simply the candidate indices.

    Note that the serialization of this layer does not preserve the candidates
    and only saves the `k` and `return_scores` arguments. One has to call
    `update_candidates` after deserializing the layers.

    Args:
        candidate_embeddings: The candidate embeddings. If `None`,
            candidates must be provided using `update_candidates` before
            using this layer.
        candidate_ids: The identifiers for the candidates. If `None` the
            indices of the candidates are returned instead.
        k: Number of candidates to retrieve.
        return_scores: When `True`, this layer returns a tuple with the top
            scores and the top identifiers. When `False`, this layer returns
            a single tensor with the top identifiers.
        **kwargs: Args to pass to the base class.

    Example:

    ```python
    retrieval = keras_rs.layers.BruteForceRetrieval(k=100)

    # At some later point, we update the candidates.
    retrieval.update_candidates(candidate_embeddings, candidate_ids)

    # We can then retrieve the top candidates for any number of queries.
    # Scores are stored highest first. Scores correspond to ids in the same row.
    tops_scores, top_ids = retrieval(query_embeddings)
    ```
    """

    def __init__(
        self,
        candidate_embeddings: Optional[types.Tensor] = None,
        candidate_ids: Optional[types.Tensor] = None,
        k: int = 10,
        return_scores: bool = True,
        **kwargs: Any,
    ) -> None:
        # Keep `k`, `return_scores` as separately passed args instead of keeping
        # them in `kwargs`. This is to ensure the user does not have to hop
        # to the base class to check which other args can be passed.
        super().__init__(k=k, return_scores=return_scores, **kwargs)

        self.candidate_embeddings = None
        self.candidate_ids = None

        if candidate_embeddings is None:
            if candidate_ids is not None:
                raise ValueError(
                    "You cannot provide `candidate_ids` without providing "
                    "`candidate_embeddings`"
                )
        else:
            self.update_candidates(candidate_embeddings, candidate_ids)

    def update_candidates(
        self,
        candidate_embeddings: types.Tensor,
        candidate_ids: Optional[types.Tensor] = None,
    ) -> None:
        """Update the set of candidates and optionally their candidate IDs.

        Args:
            candidate_embeddings: The candidate embeddings.
            candidate_ids: The identifiers for the candidates. If `None`, the
                indices of the candidates are returned instead.
        """
        self._validate_candidate_embeddings_and_ids(
            candidate_embeddings, candidate_ids
        )

        if self.candidate_embeddings is not None:
            # Update of existing variables.
            self.candidate_embeddings.assign(candidate_embeddings)

            if self.candidate_ids is None:
                if candidate_ids is not None:
                    raise ValueError(
                        "New `candidate_ids` cannot be provided as previous "
                        "candidates did not have candidate IDs"
                    )
            else:
                if candidate_ids is not None:
                    self.candidate_ids.assign(candidate_ids)
        else:
            # Initial creation of variables.
            self.candidate_embeddings = keras.Variable(
                initializer=candidate_embeddings,
                name="candidate_embeddings",
                trainable=False,
            )
            if candidate_ids is not None:
                self.candidate_ids = keras.Variable(
                    initializer=candidate_ids,
                    dtype="int32",
                    name="candidate_ids",
                    trainable=False,
                )
        self.built = True

    def call(
        self, inputs: types.Tensor
    ) -> Union[types.Tensor, tuple[types.Tensor, types.Tensor]]:
        """Returns the top candidates for the query passed as input.

        Args:
            inputs: the query for which to return top candidates.

        Returns:
            A tuple with the top scores and the top identifiers if
            `returns_scores` is True, otherwise a tensor with the top
            identifiers.
        """
        scores = self.compute_score(inputs, self.candidate_embeddings)
        top_scores, top_ids = keras.ops.top_k(scores, k=self.k)

        if self.candidate_ids is not None:
            top_ids = keras.ops.take(self.candidate_ids, top_ids, axis=0)

        if self.return_scores:
            return top_scores, top_ids
        else:
            return top_ids



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from keras_rs.src.metrics.dcg import DCG as DCG
from keras_rs.src.metrics.mean_average_precision import (
    MeanAveragePrecision as MeanAveragePrecision,
)
from keras_rs.src.metrics.mean_reciprocal_rank import (
    MeanReciprocalRank as MeanReciprocalRank,
)
from keras_rs.src.metrics.ndcg import NDCG as NDCG
from keras_rs.src.metrics.precision_at_k import PrecisionAtK as PrecisionAtK
from keras_rs.src.metrics.recall_at_k import RecallAtK as RecallAtK



from typing import Optional

from keras import ops

from keras_rs.src import types
from keras_rs.src.api_export import keras_rs_export
from keras_rs.src.losses.pairwise_loss import PairwiseLoss
from keras_rs.src.losses.pairwise_loss import pairwise_loss_subclass_doc_string
from keras_rs.src.losses.pairwise_loss_utils import apply_pairwise_op


@keras_rs_export("keras_rs.losses.PairwiseMeanSquaredError")
class PairwiseMeanSquaredError(PairwiseLoss):
    def pairwise_loss(self, pairwise_logits: types.Tensor) -> types.Tensor:
        # Since we override `compute_unreduced_loss`, we do not need to
        # implement this method.
        pass

    def compute_unreduced_loss(
        self,
        labels: types.Tensor,
        logits: types.Tensor,
        mask: Optional[types.Tensor] = None,
    ) -> tuple[types.Tensor, types.Tensor]:
        # Override `PairwiseLoss.compute_unreduced_loss` since pairwise weights
        # for MSE are computed differently.

        batch_size, list_size = ops.shape(labels)

        # Mask all values less than 0 (since less than 0 implies invalid
        # labels).
        valid_mask = ops.greater_equal(labels, ops.cast(0.0, labels.dtype))

        if mask is not None:
            valid_mask = ops.logical_and(valid_mask, mask)

        # Compute the difference for all pairs in a list. The output is a tensor
        # with shape `(batch_size, list_size, list_size)`, where `[:, i, j]`
        # stores information for pair `(i, j)`.
        pairwise_labels_diff = apply_pairwise_op(labels, ops.subtract)
        pairwise_logits_diff = apply_pairwise_op(logits, ops.subtract)
        valid_pair = apply_pairwise_op(valid_mask, ops.logical_and)
        pairwise_mse = ops.square(pairwise_labels_diff - pairwise_logits_diff)

        # Compute weights.
        pairwise_weights = ops.ones_like(pairwise_mse)
        # Exclude self pairs.
        pairwise_weights = ops.subtract(
            pairwise_weights,
            ops.tile(ops.eye(list_size, list_size), (batch_size, 1, 1)),
        )
        # Include only valid pairs.
        pairwise_weights = ops.multiply(
            pairwise_weights, ops.cast(valid_pair, dtype=pairwise_weights.dtype)
        )

        return pairwise_mse, pairwise_weights


formula = "loss = sum_{i} sum_{j} I(y_i > y_j) * (s_i - s_j)^2"
explanation = """
    - `(s_i - s_j)^2` is the squared difference between the predicted scores
      of items `i` and `j`, which penalizes discrepancies between the predicted
      order of items relative to their true order.
"""
extra_args = ""
example = """
    With `compile()` API:

    ```python
    model.compile(
        loss=keras_rs.losses.PairwiseMeanSquaredError(),
        ...
    )
    ```

    As a standalone function with unbatched inputs:

    >>> y_true = np.array([1.0, 0.0, 1.0, 3.0, 2.0])
    >>> y_pred = np.array([1.0, 3.0, 2.0, 4.0, 0.8])
    >>> pairwise_mse = keras_rs.losses.PairwiseMeanSquaredError()
    >>> pairwise_mse(y_true=y_true, y_pred=y_pred)
    >>> 19.10400

    With batched inputs using default 'auto'/'sum_over_batch_size' reduction:

    >>> y_true = np.array([[1.0, 0.0, 1.0, 3.0], [0.0, 1.0, 2.0, 3.0]])
    >>> y_pred = np.array([[1.0, 3.0, 2.0, 4.0], [1.0, 1.8, 2.0, 3.0]])
    >>> pairwise_mse = keras_rs.losses.PairwiseMeanSquaredError()
    >>> pairwise_mse(y_true=y_true, y_pred=y_pred)
    5.57999
    
    With masked inputs (useful for ragged inputs):

    >>> y_true = {
    ...     "labels": np.array([[1.0, 0.0, 1.0, 3.0], [0.0, 1.0, 2.0, 3.0]]),
    ...     "mask": np.array(
    ...         [[True, True, True, True], [True, True, False, False]]
    ...     ),
    ... }
    >>> y_pred = np.array([[1.0, 3.0, 2.0, 4.0], [1.0, 1.8, 2.0, 3.0]])
    >>> pairwise_mse(y_true=y_true, y_pred=y_pred)
    4.76000

    With `sample_weight`:

    >>> y_true = np.array([[1.0, 0.0, 1.0, 3.0], [0.0, 1.0, 2.0, 3.0]])
    >>> y_pred = np.array([[1.0, 3.0, 2.0, 4.0], [1.0, 1.8, 2.0, 3.0]])
    >>> sample_weight = np.array(
    ...     [[2.0, 3.0, 1.0, 1.0], [2.0, 1.0, 0.0, 0.0]]
    ... )
    >>> pairwise_mse = keras_rs.losses.PairwiseMeanSquaredError()
    >>> pairwise_mse(
    ...     y_true=y_true, y_pred=y_pred, sample_weight=sample_weight
    ... )
    11.0500

    Using `'none'` reduction:

    >>> y_true = np.array([[1.0, 0.0, 1.0, 3.0], [0.0, 1.0, 2.0, 3.0]])
    >>> y_pred = np.array([[1.0, 3.0, 2.0, 4.0], [1.0, 1.8, 2.0, 3.0]])
    >>> pairwise_mse = keras_rs.losses.PairwiseMeanSquaredError(
    ...     reduction="none"
    ... )
    >>> pairwise_mse(y_true=y_true, y_pred=y_pred)
    [[11., 17.,  5.,  5.], [2.04, 1.3199998, 1.6399999, 1.6399999]]
"""

PairwiseMeanSquaredError.__doc__ = pairwise_loss_subclass_doc_string.format(
    loss_name="mean squared error",
    formula=formula,
    explanation=explanation,
    extra_args=extra_args,
    example=example,
)



from typing import Optional

from keras import ops

from keras_rs.src import types
from keras_rs.src.utils.keras_utils import check_rank
from keras_rs.src.utils.keras_utils import check_shapes_compatible


def standardize_call_inputs_ranks(
    y_true: types.Tensor,
    y_pred: types.Tensor,
    mask: Optional[types.Tensor] = None,
    check_y_true_rank: bool = True,
) -> tuple[types.Tensor, types.Tensor, Optional[types.Tensor], bool]:
    """
    Utility function for processing inputs for losses and metrics.

    This utility function does three things:

    - Checks that `y_true`, `y_pred` are of rank 1 or 2;
    - Checks that `y_true`, `y_pred`, `mask` have the same shape;
    - Adds batch dimension if rank = 1.

    Args:
        y_true: tensor. Ground truth values.
        y_pred: tensor. The predicted values.
        mask: tensor. Boolean mask for `y_true`.
        check_y_true_rank: bool. Whether to check the rank of `y_true`.

    Returns:
        Tuple of processed `y_true`, `y_pred`, `mask`, and `batched`. `batched`
        is a bool indicating if the inputs are batched.
    """

    y_true_shape = ops.shape(y_true)
    y_true_rank = len(y_true_shape)
    y_pred_shape = ops.shape(y_pred)
    y_pred_rank = len(y_pred_shape)
    if mask is not None:
        mask_shape = ops.shape(mask)
        mask_rank = len(mask_shape)

    if check_y_true_rank:
        check_rank(y_true_rank, allowed_ranks=(1, 2), tensor_name="y_true")
    check_rank(y_pred_rank, allowed_ranks=(1, 2), tensor_name="y_pred")
    if mask is not None:
        check_rank(mask_rank, allowed_ranks=(1, 2), tensor_name="mask")
    if not check_shapes_compatible(y_true_shape, y_pred_shape):
        raise ValueError(
            "`y_true` and `y_pred` should have the same shape. Received: "
            f"`y_true.shape` = {y_true_shape}, `y_pred.shape` = {y_pred_shape}."
        )
    if mask is not None and not check_shapes_compatible(
        y_true_shape, mask_shape
    ):
        raise ValueError(
            "`y_true['labels']` and `y_true['mask']` should have the same "
            f"shape. Received: `y_true['labels'].shape` = {y_true_shape}, "
            f"`y_true['mask'].shape` = {mask_shape}."
        )

    batched = True
    if y_true_rank == 1:
        batched = False

        y_true = ops.expand_dims(y_true, axis=0)
        y_pred = ops.expand_dims(y_pred, axis=0)
        if mask is not None:
            mask = ops.expand_dims(mask, axis=0)

    return y_true, y_pred, mask, batched



import abc
from typing import Any, Optional

import keras
from keras import ops

from keras_rs.src import types
from keras_rs.src.losses.pairwise_loss_utils import pairwise_comparison
from keras_rs.src.metrics.utils import standardize_call_inputs_ranks


class PairwiseLoss(keras.losses.Loss, abc.ABC):
    """Base class for pairwise ranking losses.

    Pairwise loss functions are designed for ranking tasks, where the goal is to
    correctly order items within each list. Any pairwise loss function computes
    the loss value by comparing pairs of items within each list, penalizing
    cases where an item with a higher true label has a lower predicted score
    than an item with a lower true label.

    In order to implement any kind of pairwise loss, override the
    `pairwise_loss` method.
    """

    def __init__(self, temperature: float = 1.0, **kwargs: Any) -> None:
        super().__init__(**kwargs)

        if temperature <= 0.0:
            raise ValueError(
                f"`temperature` should be a positive float. Received: "
                f"`temperature` = {temperature}."
            )

        self.temperature = temperature

        # TODO(abheesht): Add `lambda_weights`.

    @abc.abstractmethod
    def pairwise_loss(self, pairwise_logits: types.Tensor) -> types.Tensor:
        pass

    def compute_unreduced_loss(
        self,
        labels: types.Tensor,
        logits: types.Tensor,
        mask: Optional[types.Tensor] = None,
    ) -> tuple[types.Tensor, types.Tensor]:
        # Mask all values less than 0 (since less than 0 implies invalid
        # labels).
        valid_mask = ops.greater_equal(labels, ops.cast(0.0, labels.dtype))

        if mask is not None:
            valid_mask = ops.logical_and(valid_mask, mask)

        pairwise_labels, pairwise_logits = pairwise_comparison(
            labels=labels,
            logits=logits,
            mask=valid_mask,
            logits_op=ops.subtract,
        )
        pairwise_logits = ops.divide(
            pairwise_logits,
            ops.cast(self.temperature, dtype=pairwise_logits.dtype),
        )

        return self.pairwise_loss(pairwise_logits), pairwise_labels

    def call(
        self,
        y_true: types.Tensor,
        y_pred: types.Tensor,
    ) -> types.Tensor:
        """Compute the pairwise loss.

        Args:
            y_true: tensor or dict. Ground truth values. If tensor, of shape
                `(list_size)` for unbatched inputs or `(batch_size, list_size)`
                for batched inputs. If an item has a label of -1, it is ignored
                in loss computation. If it is a dictionary, it should have two
                keys: `"labels"` and `"mask"`. `"mask"` can be used to ignore
                elements in loss computation, i.e., pairs will not be formed
                with those items. Note that the final mask is an `and` of the
                passed mask, and `labels >= 0`.
            y_pred: tensor. The predicted values, of shape `(list_size)` for
                unbatched inputs or `(batch_size, list_size)` for batched
                inputs. Should be of the same shape as `y_true`.

        Returns:
            The loss.
        """
        mask = None
        if isinstance(y_true, dict):
            if "labels" not in y_true:
                raise ValueError(
                    '`"labels"` should be present in `y_true`. Received: '
                    f"`y_true` = {y_true}"
                )

            mask = y_true.get("mask", None)
            y_true = y_true["labels"]

        y_true = ops.convert_to_tensor(y_true)
        y_pred = ops.convert_to_tensor(y_pred)
        if mask is not None:
            mask = ops.convert_to_tensor(mask)

        y_true, y_pred, mask, _ = standardize_call_inputs_ranks(
            y_true, y_pred, mask
        )

        losses, weights = self.compute_unreduced_loss(
            labels=y_true, logits=y_pred, mask=mask
        )
        losses = ops.multiply(losses, weights)
        losses = ops.sum(losses, axis=-1)
        return losses

    def get_config(self) -> dict[str, Any]:
        config: dict[str, Any] = super().get_config()
        config.update({"temperature": self.temperature})
        return config


pairwise_loss_subclass_doc_string = (
    "Computes pairwise {loss_name} between true labels and predicted scores."
    """
    This loss function is designed for ranking tasks, where the goal is to
    correctly order items within each list. It computes the loss by comparing
    pairs of items within each list, penalizing cases where an item with a
    higher true label has a lower predicted score than an item with a lower
    true label.

    For each list of predicted scores `s` in `y_pred` and the corresponding list
    of true labels `y` in `y_true`, the loss is computed as follows:

    ```
    {formula}
    ```

    where:

    - `y_i` and `y_j` are the true labels of items `i` and `j`, respectively.
    - `s_i` and `s_j` are the predicted scores of items `i` and `j`,
      respectively.
    - `I(y_i > y_j)` is an indicator function that equals 1 if `y_i > y_j`,
      and 0 otherwise.{explanation}
    Args:{extra_args}
        reduction: Type of reduction to apply to the loss. In almost all cases
            this should be `"sum_over_batch_size"`. Supported options are
            `"sum"`, `"sum_over_batch_size"`, `"mean"`,
            `"mean_with_sample_weight"` or `None`. `"sum"` sums the loss,
            `"sum_over_batch_size"` and `"mean"` sum the loss and divide by the
            sample size, and `"mean_with_sample_weight"` sums the loss and
            divides by the sum of the sample weights. `"none"` and `None`
            perform no aggregation. Defaults to `"sum_over_batch_size"`.
        name: Optional name for the loss instance.
        dtype: The dtype of the loss's computations. Defaults to `None`, which
            means using `keras.backend.floatx()`. `keras.backend.floatx()` is a
            `"float32"` unless set to different value
            (via `keras.backend.set_floatx()`). If a `keras.DTypePolicy` is
            provided, then the `compute_dtype` will be utilized.

    Examples:
{example}"""
)



from typing import Callable

from keras import ops

from keras_rs.src import types


def apply_pairwise_op(
    x: types.Tensor, op: Callable[[types.Tensor, types.Tensor], types.Tensor]
) -> types.Tensor:
    return op(
        ops.expand_dims(x, axis=-1),
        ops.expand_dims(x, axis=-2),
    )


def pairwise_comparison(
    labels: types.Tensor,
    logits: types.Tensor,
    mask: types.Tensor,
    logits_op: Callable[[types.Tensor, types.Tensor], types.Tensor],
) -> tuple[types.Tensor, types.Tensor]:
    # Compute the difference for all pairs in a list. The output is a tensor
    # with shape `(batch_size, list_size, list_size)`, where `[:, i, j]` stores
    # information for pair `(i, j)`.
    pairwise_labels_diff = apply_pairwise_op(labels, ops.subtract)
    pairwise_logits = apply_pairwise_op(logits, logits_op)

    # Keep only those cases where `l_i < l_j`.
    pairwise_labels = ops.cast(
        ops.greater(pairwise_labels_diff, 0), dtype=labels.dtype
    )
    if mask is not None:
        valid_pairs = apply_pairwise_op(mask, ops.logical_and)
        pairwise_labels = ops.multiply(
            pairwise_labels, ops.cast(valid_pairs, dtype=pairwise_labels.dtype)
        )

    return pairwise_labels, pairwise_logits



from keras import ops

from keras_rs.src import types
from keras_rs.src.api_export import keras_rs_export
from keras_rs.src.metrics.ranking_metric import RankingMetric
from keras_rs.src.metrics.ranking_metric import (
    ranking_metric_subclass_doc_string,
)
from keras_rs.src.metrics.ranking_metric import (
    ranking_metric_subclass_doc_string_post_desc,
)
from keras_rs.src.metrics.ranking_metrics_utils import get_list_weights
from keras_rs.src.metrics.ranking_metrics_utils import sort_by_scores
from keras_rs.src.utils.doc_string_utils import format_docstring


@keras_rs_export("keras_rs.metrics.RecallAtK")
class RecallAtK(RankingMetric):
    def compute_metric(
        self,
        y_true: types.Tensor,
        y_pred: types.Tensor,
        mask: types.Tensor,
        sample_weight: types.Tensor,
    ) -> types.Tensor:
        # Assume: `y_true = [0, 0, 1]`, `y_pred = [0.1, 0.9, 0.2]`.
        # `sorted_y_true = [0, 1, 0]` (sorted in descending order).
        (sorted_y_true,) = sort_by_scores(
            tensors_to_sort=[y_true],
            scores=y_pred,
            mask=mask,
            k=self.k,
            shuffle_ties=self.shuffle_ties,
            seed=self.seed_generator,
        )

        relevance = ops.cast(
            ops.greater_equal(
                sorted_y_true, ops.cast(1, dtype=sorted_y_true.dtype)
            ),
            dtype=y_pred.dtype,
        )
        overall_relevance = ops.cast(
            ops.greater_equal(y_true, ops.cast(1, dtype=y_true.dtype)),
            dtype=y_pred.dtype,
        )
        per_list_recall = ops.divide_no_nan(
            ops.sum(relevance, axis=1, keepdims=True),
            ops.sum(overall_relevance, axis=1, keepdims=True),
        )

        # Get weights.
        per_list_weights = get_list_weights(
            weights=sample_weight, relevance=overall_relevance
        )

        return per_list_recall, per_list_weights


concept_sentence = (
    "It measures the proportion of relevant items found in the top-k "
    "recommendations out of the total number of relevant items for a user"
)
relevance_type = "binary indicators (0 or 1) of relevance"
score_range_interpretation = (
    "Scores range from 0 to 1, with 1 indicating that all relevant items "
    "for the user were found within the top-k recommendations"
)
formula = """```
R@k(y, s) = sum_i I[rank(s_i) < k] y_i / sum_j y_j
```

where `y_i` is the relevance label (0/1) of the item ranked at position
`i`, `I[condition]` is 1 if the condition is met, otherwise 0."""
extra_args = ""
example = """
    >>> batch_size = 2
    >>> list_size = 5
    >>> labels = np.random.randint(0, 2, size=(batch_size, list_size))
    >>> scores = np.random.random(size=(batch_size, list_size))
    >>> metric = keras_rs.metrics.RecallAtK()(
    ...     y_true=labels, y_pred=scores
    ... )

    Mask certain elements (can be used for uneven inputs):

    >>> batch_size = 2
    >>> list_size = 5
    >>> labels = np.random.randint(0, 2, size=(batch_size, list_size))
    >>> scores = np.random.random(size=(batch_size, list_size))
    >>> mask = np.random.randint(0, 2, size=(batch_size, list_size), dtype=bool)
    >>> metric = keras_rs.metrics.RecallAtK()(
    ...     y_true={"labels": labels, "mask": mask}, y_pred=scores
    ... )
"""

RecallAtK.__doc__ = format_docstring(
    ranking_metric_subclass_doc_string,
    width=80,
    metric_name="Recall@k",
    metric_abbreviation="R@k",
    concept_sentence=concept_sentence,
    relevance_type=relevance_type,
    score_range_interpretation=score_range_interpretation,
    formula=formula,
) + ranking_metric_subclass_doc_string_post_desc.format(
    extra_args=extra_args, example=example
)



import abc
from typing import Any, Optional, Union

import keras
from keras import ops

from keras_rs.src import types
from keras_rs.src.metrics.utils import standardize_call_inputs_ranks
from keras_rs.src.utils.keras_utils import check_rank
from keras_rs.src.utils.keras_utils import check_shapes_compatible


class RankingMetric(keras.metrics.Mean, abc.ABC):
    """Base class for ranking evaluation metrics (e.g., MAP, MRR, DCG, nDCG).

    Ranking metrics are used to evaluate the quality of ranked lists produced
    by a ranking model. The primary goal in ranking tasks is to order items
    according to their relevance or utility for a given query or context.
    These metrics provide quantitative measures of how well a model achieves
    this goal, typically by comparing the predicted order of items against the
    ground truth relevance judgments for each list.

    To define your own ranking metric, subclass this class and implement the
    `compute_metric` method.

    Args:
        k: int. The number of top-ranked items to consider (the 'k' in 'top-k').
            Must be a positive integer.
        shuffle_ties: bool. Whether to randomly shuffle scores before sorting.
            This is done to break ties. Defaults to `True`.
        seed: int. Random seed used for shuffling.
        name: Optional name for the loss instance.
        dtype: The dtype of the metric's computations. Defaults to `None`, which
            means using `keras.backend.floatx()`. `keras.backend.floatx()` is a
            `"float32"` unless set to different value
            (via `keras.backend.set_floatx()`). If a `keras.DTypePolicy` is
            provided, then the `compute_dtype` will be utilized.
    """

    def __init__(
        self,
        k: Optional[int] = None,
        shuffle_ties: bool = True,
        seed: Optional[Union[int, keras.random.SeedGenerator]] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)

        if k is not None and (not isinstance(k, int) or k < 1):
            raise ValueError(
                f"`k` should be a positive integer. Received: `k` = {k}."
            )

        self.k = k
        self.shuffle_ties = shuffle_ties
        self.seed = seed

        # Define `SeedGenerator`. JAX doesn't work, otherwise.
        self.seed_generator = keras.random.SeedGenerator(seed)

    @abc.abstractmethod
    def compute_metric(
        self,
        y_true: types.Tensor,
        y_pred: types.Tensor,
        mask: types.Tensor,
        sample_weight: types.Tensor,
    ) -> types.Tensor:
        """Abstract method, should be implemented by subclasses."""
        pass

    def update_state(
        self,
        y_true: types.Tensor,
        y_pred: types.Tensor,
        sample_weight: Optional[types.Tensor] = None,
    ) -> None:
        """
        Accumulates statistics for the ranking metric.

        Args:
            y_true: tensor or dict. Ground truth values. If tensor, of shape
                `(list_size)` for unbatched inputs or `(batch_size, list_size)`
                for batched inputs. If an item has a label of -1, it is ignored
                in loss computation. If it is a dictionary, it should have two
                keys: `"labels"` and `"mask"`. `"mask"` can be used to ignore
                elements in metric computation, i.e., pairs will not be formed
                with those items. Note that the final mask is an `and` of the
                passed mask, `labels >= 0`, and `sample_weight > 0`.
            y_pred: tensor. The predicted values, of shape `(list_size)` for
                unbatched inputs or `(batch_size, list_size)` for batched
                inputs. Should be of the same shape as `y_true`.
            sample_weight: float/tensor. Can be float value, or tensor of
                shape `(list_size)` or `(batch_size, list_size)`. Defaults to
                `None`.
        """
        # === Process `y_true`, if dict ===
        passed_mask = None
        if isinstance(y_true, dict):
            if "labels" not in y_true:
                raise ValueError(
                    '`"labels"` should be present in `y_true`. Received: '
                    f"`y_true` = {y_true}"
                )

            passed_mask = y_true.get("mask", None)
            y_true = y_true["labels"]

        # === Convert to tensors, if list ===
        # TODO (abheesht): Figure out if we need to cast tensors to
        # `self.dtype`.
        y_true = ops.convert_to_tensor(y_true)
        y_pred = ops.convert_to_tensor(y_pred)
        if sample_weight is not None:
            sample_weight = ops.convert_to_tensor(sample_weight)
        if passed_mask is not None:
            passed_mask = ops.convert_to_tensor(passed_mask)

        # Cast to the correct dtype.
        y_true = ops.cast(y_true, dtype=self.dtype)
        y_pred = ops.cast(y_pred, dtype=self.dtype)
        if sample_weight is not None:
            sample_weight = ops.cast(sample_weight, dtype=self.dtype)

        # === Process `sample_weight` ===
        if sample_weight is None:
            sample_weight = ops.cast(1, dtype=y_pred.dtype)

        y_true_shape = ops.shape(y_true)
        y_true_rank = len(y_true_shape)
        sample_weight_shape = ops.shape(sample_weight)
        sample_weight_rank = len(sample_weight_shape)

        # Check `y_true_rank` first. Can be 1 for unbatched inputs, 2 for
        # batched.
        check_rank(y_true_rank, allowed_ranks=(1, 2), tensor_name="y_true")

        # Check `sample_weight` rank. Should be between 0 and `y_true_rank`.
        check_rank(
            sample_weight_rank,
            allowed_ranks=tuple(range(y_true_rank + 1)),
            tensor_name="sample_weight",
        )

        if y_true_rank == 2:
            # If `sample_weight` rank is 1, it should be of shape
            # `(batch_size,)`. Otherwise, it should be of shape
            # `(batch_size, list_size)`.
            if sample_weight_rank == 1:
                check_shapes_compatible(sample_weight_shape, (y_true_shape[0],))
                # Uprank this, so that we get per-list weights here.
                sample_weight = ops.expand_dims(sample_weight, axis=1)
            elif sample_weight_rank == 2:
                check_shapes_compatible(sample_weight_shape, y_true_shape)

        # Reshape `sample_weight` to the shape of `y_true`.
        sample_weight = ops.multiply(ops.ones_like(y_true), sample_weight)

        # Mask all values less than 0 (since less than 0 implies invalid
        # labels).
        valid_mask = ops.greater_equal(y_true, ops.cast(0, y_true.dtype))
        if passed_mask is not None:
            valid_mask = ops.logical_and(valid_mask, passed_mask)

        # === Process inputs - shape checking, upranking, etc. ===
        y_true, y_pred, valid_mask, batched = standardize_call_inputs_ranks(
            y_true=y_true,
            y_pred=y_pred,
            mask=valid_mask,
            check_y_true_rank=False,
        )

        # Uprank sample_weight if unbatched.
        if not batched:
            sample_weight = ops.expand_dims(sample_weight, axis=0)

        # Get `mask` from `sample_weight`.
        sample_weight_mask = ops.greater(
            sample_weight, ops.cast(0, dtype=sample_weight.dtype)
        )
        mask = ops.logical_and(valid_mask, sample_weight_mask)

        # === Update "invalid" `y_true`, `y_pred` entries based on mask ===

        # `y_true`: assign 0 for invalid entries
        y_true = ops.where(mask, y_true, ops.zeros_like(y_true))
        # `y_pred`: assign a value slightly smaller than the smallest value
        # so that it gets sorted last.
        y_pred = ops.where(
            mask,
            y_pred,
            -keras.config.epsilon() * ops.ones_like(y_pred)
            + ops.amin(y_pred, axis=1, keepdims=True),
        )
        sample_weight = ops.where(
            mask, sample_weight, ops.cast(0, sample_weight.dtype)
        )

        # === Actual computation ===
        per_list_metric_values, per_list_metric_weights = self.compute_metric(
            y_true=y_true, y_pred=y_pred, mask=mask, sample_weight=sample_weight
        )

        # Chain to `super()` to get mean metric.
        # TODO(abheesht): Figure out if we want to return unaggregated metric
        # values too of shape `(batch_size,)` from `result()`.
        super().update_state(
            per_list_metric_values, sample_weight=per_list_metric_weights
        )

    def get_config(self) -> dict[str, Any]:
        config: dict[str, Any] = super().get_config()
        config.update(
            {"k": self.k, "shuffle_ties": self.shuffle_ties, "seed": self.seed}
        )
        return config


ranking_metric_subclass_doc_string = """
Computes {metric_name} ({metric_abbreviation}).

This metric evaluates ranking quality. {concept_sentence}. The metric processes
true relevance labels in `y_true` ({relevance_type}) against predicted scores in
`y_pred`. The scores in `y_pred` are used to determine the rank order of items,
by sorting in descending order. {score_range_interpretation}.

For each list of predicted scores `s` in `y_pred` and the corresponding list
of true labels `y` in `y_true`, the per-query {metric_abbreviation} score is
calculated as follows:
{formula}

The final {metric_abbreviation} score reported is typically the weighted
average of these per-query scores across all queries/lists in the dataset.

Note: `sample_weight` is handled differently for ranking metrics. For
batched inputs, `sample_weight` can be scalar, 1D, 2D. The scalar case and
1D case (list-wise weights) are straightforward. The 2D case (item-wise
weights) is different, in the sense that the sample weights are aggregated
to get 1D weights. For more details, refer to
`keras_rs.src.metrics.ranking_metrics_utils.get_list_weights`.
"""

ranking_metric_subclass_doc_string_post_desc = """

    Args:{extra_args}
        k: int. The number of top-ranked items to consider (the 'k' in 'top-k').
            Must be a positive integer.
        shuffle_ties: bool. Whether to randomly shuffle scores before sorting.
            This is done to break ties. Defaults to `True`.
        seed: int. Random seed used for shuffling.
        name: Optional name for the loss instance.
        dtype: The dtype of the metric's computations. Defaults to `None`, which
            means using `keras.backend.floatx()`. `keras.backend.floatx()` is a
            `"float32"` unless set to different value
            (via `keras.backend.set_floatx()`). If a `keras.DTypePolicy` is
            provided, then the `compute_dtype` will be utilized.

    Example:
{example}
"""



from typing import Any, Callable, Optional

from keras import ops
from keras.saving import deserialize_keras_object
from keras.saving import serialize_keras_object

from keras_rs.src import types
from keras_rs.src.api_export import keras_rs_export
from keras_rs.src.metrics.ranking_metric import RankingMetric
from keras_rs.src.metrics.ranking_metric import (
    ranking_metric_subclass_doc_string,
)
from keras_rs.src.metrics.ranking_metric import (
    ranking_metric_subclass_doc_string_post_desc,
)
from keras_rs.src.metrics.ranking_metrics_utils import compute_dcg
from keras_rs.src.metrics.ranking_metrics_utils import default_gain_fn
from keras_rs.src.metrics.ranking_metrics_utils import default_rank_discount_fn
from keras_rs.src.metrics.ranking_metrics_utils import get_list_weights
from keras_rs.src.metrics.ranking_metrics_utils import sort_by_scores
from keras_rs.src.utils.doc_string_utils import format_docstring


@keras_rs_export("keras_rs.metrics.DCG")
class DCG(RankingMetric):
    def __init__(
        self,
        k: Optional[int] = None,
        gain_fn: Callable[[types.Tensor], types.Tensor] = default_gain_fn,
        rank_discount_fn: Callable[
            [types.Tensor], types.Tensor
        ] = default_rank_discount_fn,
        **kwargs: Any,
    ) -> None:
        super().__init__(k=k, **kwargs)

        self.gain_fn = gain_fn
        self.rank_discount_fn = rank_discount_fn

    def compute_metric(
        self,
        y_true: types.Tensor,
        y_pred: types.Tensor,
        mask: types.Tensor,
        sample_weight: types.Tensor,
    ) -> types.Tensor:
        sorted_y_true, sorted_weights = sort_by_scores(
            tensors_to_sort=[y_true, sample_weight],
            scores=y_pred,
            k=self.k,
            mask=mask,
            shuffle_ties=self.shuffle_ties,
            seed=self.seed_generator,
        )

        dcg = compute_dcg(
            y_true=sorted_y_true,
            sample_weight=sorted_weights,
            gain_fn=self.gain_fn,
            rank_discount_fn=self.rank_discount_fn,
        )

        per_list_weights = get_list_weights(
            weights=sample_weight, relevance=self.gain_fn(y_true)
        )
        # Since we have already multiplied with `sample_weight`, we need to
        # divide by `per_list_weights` so as to nullify the multiplication
        # which `keras.metrics.Mean` will do.
        per_list_dcg = ops.divide_no_nan(dcg, per_list_weights)

        return per_list_dcg, per_list_weights

    def get_config(self) -> dict[str, Any]:
        config: dict[str, Any] = super().get_config()
        config.update(
            {
                "gain_fn": serialize_keras_object(self.gain_fn),
                "rank_discount_fn": serialize_keras_object(
                    self.rank_discount_fn
                ),
            }
        )
        return config

    @classmethod
    def from_config(cls, config: dict[str, Any]) -> "DCG":
        config["gain_fn"] = deserialize_keras_object(config["gain_fn"])
        config["rank_discount_fn"] = deserialize_keras_object(
            config["rank_discount_fn"]
        )
        return cls(**config)


concept_sentence = (
    "It computes the sum of the graded relevance scores of items, applying a "
    "configurable discount based on position"
)
relevance_type = (
    "graded relevance scores (non-negative numbers where higher values "
    "indicate greater relevance)"
)
score_range_interpretation = (
    "Scores are non-negative, with higher values indicating better ranking "
    "quality (highly relevant items are ranked higher). The score for a single "
    "list is not bounded or normalized, i.e., it does not lie in a range"
)

formula = """
```
DCG@k(y', w') = sum_{i=1}^{k} (gain_fn(y'_i) / rank_discount_fn(i))
```

where:
- `y'_i` is the true relevance score of the item ranked at position `i`
  (obtained by sorting `y_true` according to `y_pred`).
- `gain_fn` is the user-provided function mapping relevance `y'_i` to a
  gain value. The default function (`default_gain_fn`) is typically
  equivalent to `lambda y: 2**y - 1`.
- `rank_discount_fn` is the user-provided function mapping rank `i`
  to a discount value. The default function (`default_rank_discount_fn`)
  is typically equivalent to `lambda rank: 1 / log2(rank + 1)`.
- The final result aggregates these per-list scores."""
extra_args = """
        gain_fn: callable. Maps relevance scores (`y_true`) to gain values. The
            default implements `2**y - 1`.
        rank_discount_fn: function. Maps rank positions to discount
            values. The default (`default_rank_discount_fn`) implements
            `1 / log2(rank + 1)`."""
example = """
    >>> batch_size = 2
    >>> list_size = 5
    >>> labels = np.random.randint(0, 3, size=(batch_size, list_size))
    >>> scores = np.random.random(size=(batch_size, list_size))
    >>> metric = keras_rs.metrics.DCG()(
    ...     y_true=labels, y_pred=scores
    ... )

    Mask certain elements (can be used for uneven inputs):

    >>> batch_size = 2
    >>> list_size = 5
    >>> labels = np.random.randint(0, 3, size=(batch_size, list_size))
    >>> scores = np.random.random(size=(batch_size, list_size))
    >>> mask = np.random.randint(0, 2, size=(batch_size, list_size), dtype=bool)
    >>> metric = keras_rs.metrics.DCG()(
    ...     y_true={"labels": labels, "mask": mask}, y_pred=scores
    ... )
"""

DCG.__doc__ = format_docstring(
    ranking_metric_subclass_doc_string,
    width=80,
    metric_name="Discounted Cumulative Gain",
    metric_abbreviation="DCG",
    concept_sentence=concept_sentence,
    relevance_type=relevance_type,
    score_range_interpretation=score_range_interpretation,
    formula=formula,
) + ranking_metric_subclass_doc_string_post_desc.format(
    extra_args=extra_args, example=example
)



import re
import textwrap
from typing import Any


def format_docstring(template: str, width: int = 80, **kwargs: Any) -> str:
    """Formats and wraps a docstring using dedent and fill."""
    base_indent_str = " " * 4

    # Initial format
    formatted = template.format(**kwargs)

    # Dedent the whole block
    dedented_all = textwrap.dedent(formatted).strip()

    # Split into logical paragraphs/blocks.
    blocks = re.split(r"(\n\s*\n)", dedented_all)

    processed_output = []

    for block in blocks:
        stripped_block = block.strip()
        if not stripped_block:
            processed_output.append(block)
            continue

        if "```" in stripped_block:
            formula_dedented = textwrap.dedent(stripped_block)
            processed_output.append(
                textwrap.indent(formula_dedented, base_indent_str)
            )
        elif "where:" in stripped_block:
            # Expect this to be already indented.
            splitted_block = stripped_block.split("\n")
            processed_output.append(
                textwrap.indent(
                    splitted_block[0] + "\n\n" + "\n".join(splitted_block[1:]),
                    base_indent_str,
                )
            )
        else:
            processed_output.append(
                textwrap.fill(
                    stripped_block,
                    width=width - len(base_indent_str),
                    initial_indent=base_indent_str,
                    subsequent_indent=base_indent_str,
                )
            )

    final_string = "".join(processed_output).strip()
    final_string = base_indent_str + final_string
    return final_string



import abc
from typing import Any, Optional, Union

import keras

from keras_rs.src import types
from keras_rs.src.api_export import keras_rs_export


@keras_rs_export("keras_rs.layers.Retrieval")
class Retrieval(keras.layers.Layer, abc.ABC):
    """Retrieval base abstract class.

    This layer provides a common interface for all retrieval layers. In order
    to implement a custom retrieval layer, this abstract class should be
    subclassed.

    Args:
        k: int. Number of candidates to retrieve.
        return_scores: bool. When `True`, this layer returns a tuple with the
            top scores and the top identifiers. When `False`, this layer returns
            a single tensor with the top identifiers.
    """

    def __init__(
        self,
        k: int = 10,
        return_scores: bool = True,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.k = k
        self.return_scores = return_scores

    def _validate_candidate_embeddings_and_ids(
        self,
        candidate_embeddings: types.Tensor,
        candidate_ids: Optional[types.Tensor] = None,
    ) -> None:
        """Validates inputs to `update_candidates()`."""

        if candidate_embeddings is None:
            raise ValueError("`candidate_embeddings` is required.")

        if len(candidate_embeddings.shape) != 2:
            raise ValueError(
                "`candidate_embeddings` must be a tensor of rank 2 "
                "(num_candidates, embedding_size), received "
                "`candidate_embeddings` with shape "
                f"{candidate_embeddings.shape}"
            )

        if candidate_embeddings.shape[0] < self.k:
            raise ValueError(
                "The number of candidates provided "
                f"({candidate_embeddings.shape[0]}) is less than the number of "
                f"candidates to retrieve (k={self.k})."
            )

        if (
            candidate_ids is not None
            and candidate_ids.shape[0] != candidate_embeddings.shape[0]
        ):
            raise ValueError(
                "The `candidate_embeddings` and `candidate_is` tensors must "
                "have the same number of rows, got tensors of shape "
                f"{candidate_embeddings.shape} and {candidate_ids.shape}."
            )

    @abc.abstractmethod
    def update_candidates(
        self,
        candidate_embeddings: types.Tensor,
        candidate_ids: Optional[types.Tensor] = None,
    ) -> None:
        """Update the set of candidates and optionally their candidate IDs.

        Args:
            candidate_embeddings: The candidate embeddings.
            candidate_ids: The identifiers for the candidates. If `None`, the
                indices of the candidates are returned instead.
        """
        pass

    @abc.abstractmethod
    def call(
        self, inputs: types.Tensor
    ) -> Union[types.Tensor, tuple[types.Tensor, types.Tensor]]:
        """Returns the top candidates for the query passed as input.

        Args:
            inputs: the query for which to return top candidates.

        Returns:
            A tuple with the top scores and the top identifiers if
            `returns_scores` is True, otherwise a tensor with the top
            identifiers.
        """
        pass

    def compute_score(
        self, query_embedding: types.Tensor, candidate_embedding: types.Tensor
    ) -> types.Tensor:
        """Computes the standard dot product score from queries and candidates.

        Args:
            query_embedding: Tensor of query embedding corresponding to the
                queries for which to retrieve top candidates.
            candidate_embedding: Tensor of candidate embeddings.

        Returns:
            The dot product of queries and candidates.
        """

        return keras.ops.matmul(
            query_embedding, keras.ops.transpose(candidate_embedding)
        )

    def get_config(self) -> dict[str, Any]:
        config: dict[str, Any] = super().get_config()
        config.update(
            {
                "k": self.k,
                "return_scores": self.compute_score,
            }
        )
        return config



import keras
import ml_dtypes
from keras import ops

from keras_rs.src import types
from keras_rs.src.api_export import keras_rs_export
from keras_rs.src.utils import keras_utils

SMALLEST_FLOAT = ml_dtypes.finfo("float32").smallest_normal / 100.0


@keras_rs_export("keras_rs.layers.RemoveAccidentalHits")
class RemoveAccidentalHits(keras.layers.Layer):
    """Zeroes the logits of accidental negatives.

    Zeroes the logits of negative candidates that have the same ID as the
    positive candidate in that row.

    Example:

    ```python
    # Create layer with the configured number of hard negatives to mine.
    remove_accidental_hits = keras_rs.layers.RemoveAccidentalHits()

    # This will zero the logits of negative candidates that have the same ID as
    # the positive candidate from `labels` so as to not negatively impact the
    # true positive.
    logits = remove_accidental_hits(logits, labels, candidate_ids)
    ```
    """

    def call(
        self,
        logits: types.Tensor,
        labels: types.Tensor,
        candidate_ids: types.Tensor,
    ) -> types.Tensor:
        """Zeroes selected logits.

        For each row in the batch, zeroes the logits of negative candidates that
        have the same ID as the positive candidate in that row.

        Args:
            logits: The logits tensor, typically `[batch_size, num_candidates]`
                but can have more dimensions or be 1D as `[num_candidates]`.
            labels: The one-hot labels tensor, must be the same shape as
                `logits`.
            candidate_ids: The candidate identifiers tensor, can be
                `[num_candidates]` or `[batch_size, num_candidates]` or have
                more dimensions as long as they match the last dimensions of
                `labels`.

        Returns:
            The modified logits with the same shape as the input logits.
        """
        # A more principled way is to implement
        # `softmax_cross_entropy_with_logits` with a input mask. Here we
        # approximate so by letting accidental hits have extremely small logits
        # (SMALLEST_FLOAT) for ease-of-implementation.

        labels_shape = ops.shape(labels)
        labels_rank = len(labels_shape)
        logits_shape = ops.shape(logits)
        candidate_ids_shape = ops.shape(candidate_ids)
        candidate_ids_rank = len(candidate_ids_shape)

        if not keras_utils.check_shapes_compatible(labels_shape, logits_shape):
            raise ValueError(
                "`labels` and `logits` should have the same shape. Received: "
                f"`labels.shape` = {labels_shape}, "
                f"`logits.shape` = {logits_shape}."
            )

        if not keras_utils.check_shapes_compatible(
            labels_shape[-candidate_ids_rank:], candidate_ids_shape
        ):
            raise ValueError(
                "`candidate_ids` should have the same shape as the last "
                "dimensions of `labels`. Received: "
                f"`candidate_ids.shape` = {candidate_ids_shape}, "
                f"`labels.shape` = {labels_shape}."
            )

        # Add dimensions to `candidate_ids` to have the same rank as `labels`.
        if candidate_ids_rank < labels_rank:
            candidate_ids = ops.expand_dims(
                candidate_ids, list(range(labels_rank - candidate_ids_rank))
            )
        positive_indices = ops.expand_dims(ops.argmax(labels, axis=-1), -1)
        positive_candidate_ids = ops.take(candidate_ids, positive_indices)

        duplicate = ops.cast(
            ops.equal(positive_candidate_ids, candidate_ids), labels.dtype
        )
        duplicate = ops.subtract(duplicate, labels)

        return ops.add(logits, ops.multiply(duplicate, SMALLEST_FLOAT))



from typing import Union

import keras

from keras_rs.src import types


def clone_initializer(
    initializer: Union[str, keras.initializers.Initializer],
) -> keras.initializers.Initializer:
    """Clones an initializer to ensure a new seed.

    Args:
        initializer: The initializer to clone.

    Returns:
        A cloned initializer if it is clonable, otherwise the original one.

    As of tensorflow 2.10, we need to clone user passed initializers when
    invoking them twice to avoid creating the same randomized initialization.
    """
    if isinstance(initializer, keras.initializers.Initializer):
        config = initializer.get_config()
        initializer_class: type[keras.initializers.Initializer] = (
            initializer.__class__
        )
        return initializer_class.from_config(config)
    # If we get a string or dict, just return as we cannot and should not clone.
    return initializer


def check_shapes_compatible(
    shape1: types.TensorShape, shape2: types.TensorShape
) -> bool:
    # Check rank first.
    if len(shape1) != len(shape2):
        return False

    for d1, d2 in zip(shape1, shape2):
        if isinstance(d1, int) and isinstance(d2, int):
            if d1 != d2:
                return False

    return True


def check_rank(
    x_rank: int,
    allowed_ranks: tuple[int, ...],
    tensor_name: str,
) -> None:
    if x_rank not in allowed_ranks:
        raise ValueError(
            f"`{tensor_name}` should have a rank from `{allowed_ranks}`."
            f"Received: `{x_rank}`."
        )



from typing import Any, TypeVar

import keras

try:
    import namex
except ImportError:
    namex = None


T = TypeVar("T", bound=Any)


class keras_rs_export:
    def __init__(self, path: str):
        if namex is not None:
            self.namex_export = namex.export(package="keras_rs", path=path)

    def __call__(self, symbol: T) -> T:
        keras.saving.register_keras_serializable(package="keras_rs")(symbol)
        if namex is not None:
            self.namex_export(symbol)
        return symbol



"""Type definitions."""

from typing import Any, Optional, Sequence

"""
A tensor in any of the backends.

We do not define it explicitly to not require all the backends to be installed
and imported. The explicit definition would be:
```
Union[
  numpy.ndarray,
  tensorflow.Tensor,
  tensorflow.RaggedTensor,
  tensorflow.SparseTensor,
  tensorflow.IndexedSlices,
  jax.Array,
  jax.experimental.sparse.JAXSparse,
  torch.Tensor,
  keras.KerasTensor,
]
```
"""
Tensor = Any

TensorShape = Sequence[Optional[int]]
